"""Functions to convert objects between PyAmf and AmFast."""

import pyamf
import pyamf.remoting as pyamf_remoting
import pyamf.flex.messaging as pyamf_messaging

import amfast
import amfast.class_def as class_def
import amfast.remoting as amfast_remoting
import amfast.remoting.flex_messages as amfast_messaging

class PyAmfConversionError(amfast.AmFastError):
    """Raised when conversion to/from PyAmf datatype fails."""
    pass

class PyAmfVersionError(PyAmfConversionError):
    """Raised when installed version of PyAmf is not compatible."""

if pyamf.__version__[1] < 4:
    raise PyAmfVersionError('PyAmf version is not compatible.')

#---------- FROM PyAmf TO AmFast -----------#
def packet_to_amfast(pyamf_packet):
    """Converts a PyAmf Envelope to an AmFast Packet.

    arguments:
    ===========
     * pyamf_packet - pyamf.remoting.Envelope.

    Returns amfast.remoting.Packet
    """
    if pyamf_packet.clientType == pyamf.ClientTypes.Flash6:
        client_type = amfast_remoting.Packet.FLASH_8
    elif pyamf_packet.clientType == pyamf.ClientTypes.FlashCom:
        client_type = amfast_remoting.Packet.FLASH_COM
    elif pyamf_packet.clientType == pyamf.ClientTypes.Flash9:
        client_type = amfast_remoting.Packet.FLASH_9
    else:
        clientType = amfast_remoting.Packet.FLASH_8 

    headers = [amfast_remoting.Header(name,
        required=pyamf_packet.headers.is_required(name), value=header) \
        for name, header in pyamf_packet.headers]

    messages = [message_to_amfast(name, body) for name, body in pyamf_packet.bodies]

    return amfast_remoting.Packet(client_type=client_type, headers=headers, messages=messages)

def message_to_amfast(name, msg):
    """Converts a PyAmf Request to an AmFast Message

    arguments:
    ===========
     * name - string, response
     * msg - pyamf.remoting.Request

    Returns amfast.remoting.Message
    """ 
    if hasattr(msg, 'target'):
        target = msg.target
    else:
        target = ''

    if hasattr(msg, 'status'):
        response = name + msg.status
    else:
        response = name

    return amfast_remoting.Message(target=target, response=response, body=msg.body)

#--------- FROM AmFast to PyAmf -----------#
def dummy_callable(obj):
    """A callable that you probably shouldn't be using :)"""
    return []

def class_def_alias(class_def):
    """Create a pyamf.ClassAlias object that uses a ClassDef for the actual operations.

    arguments:
    ==========
     * class_def - amfast.class_def.ClassDef

    Returns pyamf.ClassAlias
    """
    metadata = []
    if hasattr(class_def, 'DYNAMIC_CLASS_DEF'):
        metadata.append('dynamic')
    elif hasattr(class_def, 'EXTERNALIZABLE_CLASS_DEF'):
        metadata.append('external')
    else:
        metadata.append('static')

    if class_def.amf3 is True:
        metadata.append('amf3')

    class_alias = ClassDefAlias(class_def.class_, class_def.alias,
        attrs=class_def.static_attrs, attr_func=dummy_callable,
        metadata=metadata)
    class_alias.class_def = class_def
    return class_alias

def register_class_def(class_def):
    """Maps a ClassDef to PyAmf.

    This provides the functionality of pyamf.register_class,
    except it maps a ClassDef.

    arguments:
    ===========
     * class_def - amfast.class_def.ClassDef
    """
    if class_def.alias in pyamf.CLASS_CACHE:
        raise PyAmfConversionError("Alias '%s' is already registered." % class_def.alias)

    class_alias = class_def_alias(class_def)
    if pyamf.__version__[1] > 4:
        pyamf.CLASS_CACHE[class_alias.klass] = class_alias
    pyamf.CLASS_CACHE[class_alias.alias] = class_alias

def register_class_mapper(class_mapper):
    """Maps all ClassDefs in a ClassDefMapper to PyAmf.

    arguments:
    ===========
     * class_mapper - amfast.class_def.ClassDefMapper
    """

    for class_def in class_mapper:
        if class_def._built_in is False:
            register_class_def(class_def)

def packet_to_pyamf(amfast_packet):
    """Converts an AmFast Packet to a PyAmf Envelope

    arguments:
    ==========
     * amfast.remoting.Packet

    Returns pyamf.remoting.Evenlope
    """

    version = pyamf.AMF0

    if amfast_packet.client_type == amfast_remoting.Packet.FLASH_8:
        client_type = pyamf.ClientTypes.Flash6
    elif amfast_packet.client_type == amfast_remoting.Packet.FLASH_COM:
        client_type = pyamf.ClientTypes.FlashCom
    elif amfast_packet.client_type == amfast_remoting.Packet.FLASH_9:
        client_type = pyamf.ClientTypes.Flash9
    else:
        client_type = pyamf.ClientTypes.Flash6

    packet = pyamf_remoting.Envelope()
    packet.amfVersion = version
    packet.clientType = client_type

    headers = pyamf_remoting.HeaderCollection()
    for header in amfast_packet.headers:
        headers[header.name] = header.value
        if header.required is True:
            headers.set_required(header.name, value=True)
        else:
            headers.set_required(header.name, value=False)
    packet.headers = headers

    for msg in amfast_packet.messages:
        split_target = msg.target.split('/')
        pyamf_status = '/' + split_target.pop()
        pyamf_target = '/'.join(split_target)
        packet[pyamf_target] = message_to_pyamf(msg, packet, pyamf_status) 

    return packet

def message_to_pyamf(msg, packet, status):
    """Converts an AmFast Message to a PyAmf Response.

    arguments:
    ===========
     * msg - amfast.remoting.Message
     * packet - pyamf.remoting.Envelope
     * status - string

    Returns pyamf.remoting.Response
    """

    message = pyamf_remoting.Response(msg.body)
    message.envelope = packet

    for k, v in pyamf_remoting.STATUS_CODES.iteritems():
        if v == status:
            message.status = k
            break

    return message

#----------- PyAMF Class Extensions ------------#
# Some extra classes to smooth things along with AmFast.

class ClassDefAlias(pyamf.ClassAlias):
    """A pyamf.ClassAlias that uses an amfast.class_def.ClassDef
    on the backend. This class should be instaniated with the
    class_def_alias() function.
    """

    def checkClass(kls, klass):
        # Override parent method, because
        # AmFast does not require that mapped
        # classes' __init__ methods
        # have no required arguments.
        pass

    def getAttrs(self, obj, *args, **kwargs):
        """Returns attribute names in PyAmf format."""
        if hasattr(self.class_def, 'DYNAMIC_CLASS_DEF'):
            dynamic_attrs = self.class_def.getDynamicAttrVals(obj).keys()
        else:
            dynamic_attrs = []

        return (self.class_def.static_attrs, dynamic_attrs)

    def getAttributes(self, obj, *args, **kwargs):
        """Returns attribute values in PyAmf format."""
        if hasattr(self.class_def, 'DYNAMIC_CLASS_DEF'):
            dynamic_attrs = self.class_def.getDynamicAttrVals(obj)
        else:
            dynamic_attrs = {}

        static_attrs = {}
        static_attr_vals = self.class_def.getStaticAttrVals(obj)
        for i in xrange(0, len(self.class_def.static_attrs)):
            static_attrs[self.class_def.static_attrs[i]] = static_attr_vals[i]

        return (static_attrs, dynamic_attrs)

    def applyAttributes(self, obj, attrs, *args, **kwargs):
        """Applies attributes to an instance."""
        self.class_def.applyAttrVals(obj, attrs)

    def createInstance(self, *args, **kwargs):
        """Returns a new instance of the mapped class."""
        return self.class_def.getInstance()

#---- Classes for dealing with ISmallMessage ----#
class DataInputReader(object):
    """A wrapper class for pyamf.amf3.DataInput.
 
    Use this, so we can re-use our existing ISmallMsg reading code.
    """

    def __init__(self, data_input):
        self.data_input = data_input

    def read(self, length):
        return self.data_input.stream.read(length)

    def readElement(self):
        return self.data_input.decoder.readElement()

class PyamfAbstractSmallMsgDef(amfast_messaging.AbstractSmallMsgDef):
    """Decodes ISmallMessages with PyAmf."""

    def readExternal(self, obj, data_input):
        """Overridden to use PyAmf instead of AmFast."""
        flags = self._readFlags(data_input)

        for i, flag in enumerate(flags):
            if i == 0:
                if flag & self.BODY_FLAG:
                    obj.body = data_input.readElement()
                else:
                    obj.body = None

                if flag & self.CLIENT_ID_FLAG:
                    obj.clientId = data_input.readElement()
                else:
                    obj.clientId = None

                if flag & self.DESTINATION_FLAG:
                    obj.destination = data_input.readElement()
                else:
                   obj.destination = None

                if flag & self.HEADERS_FLAG:
                    obj.headers = data_input.readElement()
                else:
                    obj.headers = None

                if flag & self.MESSAGE_ID_FLAG:
                    obj.messageId = data_input.readElement()
                else:
                    obj.messageId = None

                if flag & self.TIMESTAMP_FLAG:
                    obj.timestamp = data_input.readElement()
                else:
                    obj.timestamp = None

                if flag & self.TIME_TO_LIVE_FLAG:
                    obj.timeToLive = data_input.readElement()
                else:
                    obj.timeToLive = None

            if i == 1:
                if flag & self.CLIENT_ID_BYTES_FLAG:
                    clientIdBytes = data_input.readElement()
                    obj.clientId = self._readUid(clientIdBytes)
                else:
                    if not hasattr(obj, 'clientId'):
                        obj.clientId = None

                if flag & self.MESSAGE_ID_BYTES_FLAG:
                    messageIdBytes = data_input.readElement()
                    obj.messageId = self._readUid(messageIdBytes)
                else:
                    if not hasattr(obj, 'messageId'):
                        obj.messageId = None

    def getInstance(self):
        """
        Return a regular AmFast AbstractMessage instead of
        the class that has been mapped to this ClassDef.

        Kinda tricky. Muuuuhhahahahah
        """
        obj = amfast_messaging.AbstractMessage.__new__(amfast_messaging.AbstractMessage)

        def readAmf(data_input):
            self.readExternal(obj, DataInputReader(data_input))
        obj.__readamf__ = readAmf
        return obj

class PyamfAsyncSmallMsgDef(amfast_messaging.AsyncSmallMsgDef, PyamfAbstractSmallMsgDef):

    def __init__(self, *args, **kwargs):
        amfast_messaging.AsyncSmallMsgDef.__init__(self, *args, **kwargs)

    def readExternal(self, obj, data_input):
        PyamfAbstractSmallMsgDef.readExternal(self, obj, data_input)

        flags = self._readFlags(data_input)
        for i, flag in enumerate(flags):
            if i == 0:
                if flag & self.CORRELATION_ID_FLAG:
                    obj.correlationId = data_input.readElement()
                else:
                    obj.correlationId = None

                if flag & self.CORRELATION_ID_BYTES_FLAG:
                    correlationIdBytes = data_input.readElement()
                    obj.correlationId = self._readUid(correlationIdBytes)
                else:
                    if not hasattr(obj, 'correlationId'):
                        obj.correlationId = None

    def getInstance(self):
        obj = amfast_messaging.AsyncMessage.__new__(amfast_messaging.AsyncMessage)

        def readAmf(data_input):
            self.readExternal(obj, DataInputReader(data_input))
        obj.__readamf__ = readAmf
        return obj

class PyamfCommandSmallMsgDef(amfast_messaging.CommandSmallMsgDef, PyamfAsyncSmallMsgDef):

    def __init__(self, *args, **kwargs):
        amfast_messaging.CommandSmallMsgDef.__init__(self, *args, **kwargs)

    def readExternal(self, obj, data_input):
        PyamfAsyncSmallMsgDef.readExternal(self, obj, data_input)

        flags = self._readFlags(data_input)
        for i, flag in enumerate(flags):
            if i == 0:
                if flag & self.OPERATION_FLAG:
                    obj.operation = data_input.readElement()
                else:
                    obj.operation = None

    def getInstance(self):
        obj = amfast_messaging.CommandMessage.__new__(amfast_messaging.CommandMessage)

        def readAmf(data_input):
            self.readExternal(obj, DataInputReader(data_input))
        obj.__readamf__ = readAmf
        return obj

# ---- Dummy classes to trick PyAmf into doing what we want. ---#
class SmallAbstractMsg(amfast_messaging.AbstractMessage):
    def __readamf__(self, data_input):
        raise pyamf.EncodeError("__readamf__ is not implemented for this class: %s." % self)

    def __writeamf__(self, data_output):
        raise pyamf.EncodeError("__writeamf__ is not implemented for this class: %s." % self)

class SmallAsyncMsg(amfast_messaging.AsyncMessage):
    def __readamf__(self, data_input):
        raise pyamf.EncodeError("__readamf__ is not implemented for this class: %s." % self)

    def __writeamf__(self, data_output):
        raise pyamf.EncodeError("__writeamf__ is not implemented for this class: %s." % self)

class SmallCommandMsg(amfast_messaging.CommandMessage):
    def __readamf__(self, data_input):
        raise pyamf.EncodeError("__readamf__ is not implemented for this class: %s." % self)

    def __writeamf__(self, data_output):
        raise pyamf.EncodeError("__writeamf__ is not implemented for this class: %s." % self)

#----- Map Flex message classes with PyAmf -----#

# Clear existing message class mappings,
# then re-map with AmFast ClassDefs.

#---- AbstractMessage ---#
try:
    pyamf.unregister_class('flex.messaging.messages.AbstractMessage')
except pyamf.UnknownClassAlias:
    pass
register_class_def(class_def.ClassDef(amfast_messaging.AbstractMessage))

#---- AsyncMessage ----#
try:
    pyamf.unregister_class('flex.messaging.messages.AsyncMessage')
except pyamf.UnknownClassAlias:
    pass

try:
    pyamf.unregister_class('DSA')
except pyamf.UnknownClassAlias:
    pass

register_class_def(class_def.ClassDef(amfast_messaging.AsyncMessage))
register_class_def(PyamfAsyncSmallMsgDef(SmallAsyncMsg, 'DSA',
    ('body', 'clientId', 'destination', 'headers', 'messageId',
        'timeToLive', 'timestamp', 'correlationId')))

#---- AcknowledgeMessage --#
try:
    pyamf.unregister_class('flex.messaging.messages.AcknowledgeMessage')
except pyamf.UnknownClassAlias:
    pass

try:
    pyamf.unregister_class('DSK')
except pyamf.UnknownClassAlias:
    pass

register_class_def(class_def.ClassDef(amfast_messaging.AcknowledgeMessage))

#---- CommandMessage ----#
try:
    pyamf.unregister_class('flex.messaging.messages.CommandMessage')
except pyamf.UnknownClassAlias:
    pass

try:
    pyamf.unregister_class('DSC')
except pyamf.UnknownClassAlias:
    pass

register_class_def(class_def.ClassDef(amfast_messaging.CommandMessage))
register_class_def(PyamfCommandSmallMsgDef(SmallCommandMsg, 'DSC',
    ('body', 'clientId', 'destination', 'headers', 'messageId',
        'timeToLive', 'timestamp', 'correlationId', 'operation')))

#---- ErrorMessage ----#
try:
    pyamf.unregister_class('flex.messaging.messages.ErrorMessage')
except pyamf.UnknownClassAlias:
    pass

register_class_def(class_def.ClassDef(amfast_messaging.ErrorMessage))


#---- RemotingMessage ----#
try:
    pyamf.unregister_class('flex.messaging.messages.RemotingMessage')
except pyamf.UnknownClassAlias:
    pass

register_class_def(class_def.ClassDef(amfast_messaging.RemotingMessage))



import unittest

import amfast
amfast.use_dummy_threading = True

def suite():
    import amf3_encoder_test
    import amf0_encoder_test
    import amf3_decoder_test
    import amf0_decoder_test
    import round_trip_test
    import remoting_test
    import buffer_test
    import context_test
    import connection_test
    import subscription_test
    import messaging_test

    return unittest.TestSuite((
        amf3_decoder_test.suite(),
        amf3_encoder_test.suite(),
        amf0_encoder_test.suite(),
        amf0_decoder_test.suite(),
        round_trip_test.suite(),
        remoting_test.suite(),
        buffer_test.suite(),
        context_test.suite(),
        connection_test.suite(),
        subscription_test.suite(),
        messaging_test.suite()
    ))

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())    



import unittest

from amfast.buffer import Buffer, BufferError

class BufferTestCase(unittest.TestCase):
    def setUp(self):
        self.test_string = 'tester'

    def testRead(self):
        buf = Buffer(self.test_string)
        self.assertEquals(self.test_string[0], buf.read(1))
        self.assertEquals(1, buf.tell())
        self.assertEquals(self.test_string[1:len(self.test_string)], buf.read(len(self.test_string) - 1 ))

    def testSeek(self):
        buf = Buffer(self.test_string)
        buf.seek(len(self.test_string) - 1)
        self.assertEquals(self.test_string[-1], buf.read(1))
        buf.seek(0)
        self.assertEquals(self.test_string[0], buf.read(1))

    def testGetValue(self):
        buf = Buffer(self.test_string)
        self.assertEquals(self.test_string, buf.getvalue())

    def testWrite(self):
        buf = Buffer()
        buf.write(self.test_string)
        self.assertEquals(self.test_string, buf.getvalue())
  
    def testWriteLong(self):
        tester = 's' * 65537
        buf = Buffer()
        buf.write(tester)
        self.assertEquals(tester, buf.getvalue())

    def testWriteNonStringRaisesException(self):
        buf = Buffer()
        self.assertRaises(BufferError, buf.write, (len(self.test_string)))

    def testReadFromWriteRaisesException(self):
        buf = Buffer()
        self.assertRaises(BufferError, buf.read, 1)

    def testWriteFromReadRaisesException(self):
        buf = Buffer(self.test_string)
        self.assertRaises(BufferError, buf.write, self.test_string)

    def testReadLongRaisesException(self):
        buf = Buffer(self.test_string)
        self.assertRaises(BufferError, buf.read, (len(self.test_string) + 1))

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(BufferTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



# -*- coding: utf-8 -*-
import unittest

from StringIO import StringIO

from amfast.context import DecoderContext
import amfast.decode as decode
from amfast.decoder import Decoder
import amfast.class_def as class_def

class Amf0DecoderTestCase(unittest.TestCase):
    class Spam(object):
        def __init__(self):
            self.spam = 'eggs'

    def setUp(self):
        self.class_mapper = class_def.ClassDefMapper()
        self.decoder = Decoder()

    def testFalse(self):
        self.assertEquals(False, decode.decode(DecoderContext('\x01\x00')))

    def testStringInput(self):
        self.assertEquals(False, decode.decode('\x01\x00'))

    def testDecoderObj(self):
        self.assertEquals(False, decode.decode('\x01\x00'))

    def testTrue(self):
        self.assertEquals(True, self.decoder.decode('\x01\x01'))

    def testNumber(self):
        tests = {
            0: '\x00\x00\x00\x00\x00\x00\x00\x00\x00',
            0.2: '\x00\x3f\xc9\x99\x99\x99\x99\x99\x9a',
            1: '\x00\x3f\xf0\x00\x00\x00\x00\x00\x00',
            42: '\x00\x40\x45\x00\x00\x00\x00\x00\x00',
            -123: '\x00\xc0\x5e\xc0\x00\x00\x00\x00\x00',
            1.23456789: '\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b'
        }

        for number, encoding in tests.iteritems():
            self.assertEquals(number, decode.decode(DecoderContext(encoding)))

    def testString(self):
        tests = {
            '': '\x02\x00\x00',
            'hello': '\x02\x00\x05hello'
        }

        for string, encoding in tests.iteritems():
            self.assertEquals(string, decode.decode(DecoderContext(encoding)))

    def testLongString(self):
        decoded = 's' * 65537
        encoded = '\x0C\x00\x01\x00\x01' + decoded
  
        self.assertEquals(decoded, decode.decode(DecoderContext(encoded)))

    def testNull(self):
        self.assertEquals(None, decode.decode(DecoderContext('\x05')))

    def testUndefined(self):
        self.assertEquals(None, decode.decode(DecoderContext('\x06')))

    def testAnonObj(self):
        encoded = '\x03' #header
        encoded += '\x00\x04spam\x02\x00\x04eggs' #values
        encoded += '\x00\x00\t' # terminator
        result = decode.decode(DecoderContext(encoded))
        self.assertEquals('eggs', result['spam'])

        result = decode.decode(DecoderContext(StringIO(encoded)))
        self.assertEquals('eggs', result['spam'])

    def testMixedArray(self):
        encoded = '\x08\x00\x00\x00\x00' # mixed array header
        encoded += '\x00\x04spam\x02\x00\x04eggs' #values
        encoded += '\x00\x00\t' # terminator

        result = decode.decode(DecoderContext(encoded))
        self.assertEquals('eggs', result['spam'])

        result = decode.decode(DecoderContext(StringIO(encoded)))
        self.assertEquals('eggs', result['spam'])

    def testArray(self):
        decoded = [0, 1, 1.23456789]
        encoded = '\x0A\x00\x00\x00\x03' # 3 element array header
        encoded += '\x00\x00\x00\x00\x00\x00\x00\x00\x00' # element 1
        encoded += '\x00\x3f\xf0\x00\x00\x00\x00\x00\x00' #element 2
        encoded += '\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b' #element 3

        result = decode.decode(DecoderContext(encoded))
        for i, obj in enumerate(decoded):
            self.assertEquals(obj, result[i])

        result = decode.decode(DecoderContext(StringIO(encoded)))
        for i, obj in enumerate(decoded):
            self.assertEquals(obj, result[i])

    def testDate(self):
        import datetime
        encoded = '\x0BBp+6!\x15\x80\x00\x00\x00'
        result = decode.decode(DecoderContext(encoded))
        self.assertEquals(2005, result.year)
        self.assertEquals(3, result.month)
        self.assertEquals(18, result.day)
        self.assertEquals(1, result.hour)
        self.assertEquals(58, result.minute)
        self.assertEquals(31, result.second)

    def testXml(self):
        import xml.dom.minidom

        encoded = '\x0F' # XML header
        encoded += '\x00\x00\x00\x55' # String header
        encoded += '<?xml version="1.0" ?><test>\n            <test_me>tester</test_me>\n           </test>' # encoded XML

        result = decode.decode(DecoderContext(encoded))
        self.assertEquals(xml.dom.minidom.Document, result.__class__)

    def testDynamicObj(self):
        self.class_mapper.mapClass(class_def.DynamicClassDef(self.Spam, 'alias.spam', ()))

        encoded = '\x10\x00\x0Aalias.spam'
        encoded += '\x00\x04spam\x02\x00\x04eggs\x00\x00\x09' # dynamic attrs

        result = decode.decode(DecoderContext(encoded, class_def_mapper=self.class_mapper))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals('eggs', result.spam)

    def testStaticObj(self):
        self.class_mapper.mapClass(class_def = class_def.ClassDef(self.Spam, 'alias.spam', ('spam')))

        encoded = '\x10\x00\x0Aalias.spam'
        encoded += '\x00\x04spam\x02\x00\x04eggs\x00\x00\x09' # dynamic attrs

        result = decode.decode(DecoderContext(encoded, class_def_mapper=self.class_mapper))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals('eggs', result.spam)

    def testReferences(self):
        encoded = '\x0A\x00\x00\x00\x04' # 3 element array header
        encoded += '\x03\x00\x04spam\x02\x00\x04eggs\x00\x00\t' # obj 1
        encoded += '\x07\x00\x01' # ref to obj 1
        encoded += '\x07\x00\x01' # ref to obj 1
        encoded += '\x07\x00\x00' # circular ref
        result = decode.decode(DecoderContext(encoded))

        self.assertEquals(4, len(result))
        self.assertEquals(result, result.pop(-1))
        for obj in result:
            self.assertEquals('eggs', obj['spam'])

    def testUnkownByteRaisesException(self):
        self.assertRaises(decode.DecodeError, decode.decode, DecoderContext('\xFF'))

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(Amf0DecoderTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())




"""Send and receive AMF messages."""
import time
import threading
import uuid

import amfast
from amfast.class_def import ClassDefMapper

import connection_manager as cm
import subscription_manager as sm
import flex_messages as messaging
import thread_pool

class ChannelError(amfast.AmFastError):
    pass

class SecurityError(ChannelError):
    pass

class ChannelFullError(ChannelError):
    pass

class Channel(object):
    """An individual channel that can send/receive messages.

    attributes
    ===========
     * name - string, Channel name.
     * endpoint - Endpoint, encodes and decodes messages.
     * max_connections - int, When the number of connections exceeds this number,
         an exception is raised when new clients attempt to connect. Set to -1
         for no limit.
    """

    def __init__(self, name, max_connections=-1, endpoint=None):
        self.name = name
        self.max_connections = max_connections
        if endpoint is None:
            from endpoint import AmfEndpoint
            endpoint = AmfEndpoint()
        self.endpoint = endpoint
       
        self._lock = amfast.mutex_cls() 
        self._channel_set = None

    def _getChannelSet(self):
        # channel_set should be considered
        # read-only outside of this class
        return self._channel_set
    channel_set = property(_getChannelSet)

    def encode(self, *args, **kwargs):
        """Encode a packet."""
        try:
            return self.endpoint.encodePacket(*args, **kwargs)
        except amfast.AmFastError, exc:
            # Not much we can do if packet is not encoded properly
            amfast.log_exc(exc)
            raise exc

    def decode(self, *args, **kwargs):
        """Decode a raw request."""
        try:
            return self.endpoint.decodePacket(*args, **kwargs)
        except amfast.AmFastError, exc:
            # Not much we can do if packet is not decoded properly
            amfast.log_exc(exc)
            raise exc

    def invoke(self, request):
        """Invoke an incoming request packet."""
        try:
            request.channel = self # so user can access channel object
            return request.invoke()
        except amfast.AmFastError, exc:
            return request.fail(exc)

    def getFlexConnection(self, flex_msg):
        """Returns a Connection object for a Flex message.

        Creates a new Connection if one does not already exist.

        arguments
        ==========
         * flex_msg - FlexMessage object.
        """
        # If header does not exist, connection does not exist.
        if not hasattr(flex_msg, 'headers') or flex_msg.headers is None:
            return self.connect()

        flex_client_id = flex_msg.headers.get(flex_msg.FLEX_CLIENT_ID_HEADER, None)
        if flex_client_id == 'nil' or flex_client_id is None:
            return self.connect()

        try:
            return self.channel_set.connection_manager.getConnection(flex_msg.headers[flex_msg.FLEX_CLIENT_ID_HEADER])
        except cm.NotConnectedError:
            return self.connect(flex_msg.headers[flex_msg.FLEX_CLIENT_ID_HEADER])

    def connect(self, connection_id=None):
        """Add a client connection to this channel.

        arguments
        ==========
         * flex_client_id - string, Flex client id.

        Returns Connection
        """
        if self.max_connections > -1 and \
            self.channel_set.connection_manager.getConnectionCount(self.name) \
            >= self.max_connections:
            raise ChannelFullError("Channel '%s' is not accepting connections." % self.name)

        return self.channel_set.connect(self, connection_id)

    def disconnect(self, connection):
        """Remove a client connection from this Channel.

        arguments
        ==========
         * connection - Connection.
        """

        self.channel_set.disconnect(connection)

class HttpChannel(Channel):
    """An individual channel that can send/receive messages over HTTP.

    attributes
    ===========
     * wait_interval - int, Number of millisecondsseconds to wait before sending
         response to client when a polling request is received. Set to -1 to
         configure channel as a long-polling channel. Default = 0
     * poll_interval - int, Number of milliseconds between message polling operations
         when channel_set.notify == False
    """

    # Content type for amf messages
    CONTENT_TYPE = 'application/x-amf'

    # These bytes need to be sent to 'kickstart'
    # streaming connections so that the client 
    # will respond to incoming message. In reality
    # this value is browser specific, but for now
    # we're just using the maximum required byte length
    # which is for IE.
    KICKSTART_BYTES = 2048

    def __init__(self, name, max_connections=-1, endpoint=None,
        wait_interval=0, poll_interval=500):

        Channel.__init__(self, name, max_connections, endpoint)

        self.wait_interval = wait_interval
        self.poll_interval = poll_interval

    def getBadMethodMsg(self):
        return "405 Method Not Allowed\n\nAMF request must use 'POST' method."

    def getBadEncodingMsg(self):
        return self.getBadRequestMsg('AMF packet could not be decoded.')

    def getBadRequestMsg(self, msg=''):
        return "400 Bad Request\n\n%s" % msg

    def getBadPageMsg(self, msg=''):
        return "404 Not Found\n\n%s" % msg

    def getBadServerMsg(self, msg=''):
        return "500 Internal Server Error\n\nAmFast server error.%s" % msg

    def waitForMessage(self, packet, message, connection):
        """Waits until a new message is published to this connection.
        Returns list of messages.
        """

        if self.channel_set.notify_connections is True:
            return self._waitForMessage(packet, message, connection)
        else:
            return self._pollForMessage(packet, message, connection)

    def _waitForMessage(self, packet, message, connection):
        """Waits until notified that a new message is available, then returns messages.

        This is blocking, and should only be used
        for Channels where each connection is a thread.

        Synchronous servers should override this method.
        """

        event = threading.Event()
        connection.setNotifyFunc(event.set)
        
        if (self.wait_interval > -1):
            event.wait(float(self.wait_interval) / 1000)

        # Event.set has been called,
        # or timeout has been reached.
        connection.unSetNotifyFunc()

        return self.channel_set.subscription_manager.pollConnection(connection)

    def _pollForMessage(self, packet, message, connection):
        """Repeatedly polls for a new message until
        message is available or wait_interval is reached.

        This is blocking, and should only be used
        for Channels where each connection is a thread.

        Synchronous servers should override this method.
        """
        # If True, don't store persistent 'last_polled' value every poll operation.
        soft_touch = hasattr(self.channel_set.connection_manager, "softTouchPolled")

        total_time = 0
        poll_secs = float(self.poll_interval) / 1000
        wait_secs = float(self.wait_interval) / 1000
        while True:
            event = threading.Event()
            event.wait(poll_secs)
            msgs = self.channel_set.subscription_manager.pollConnection(connection, soft_touch)
            if len(msgs) > 0:
                if soft_touch is True:
                    # Store 'last_polled' value.
                    connection.touchPolled()            
                return msgs
            
            total_time += poll_secs
            if total_time > wait_secs or connection.connected is False:
                if soft_touch is True:
                    # Store 'last_polled' value.
                    connection.touchPolled()
                return ()

class ChannelSet(object):
    """A collection of Channels.

    A client can access the same RPC exposed methods
    from any of the Channels contained in a ChannelSet.

    A Channel can only belong to 1 ChannelSet.

    attributes
    ===========
     * service_mapper - ServiceMapper, maps destinations to Targets.
     * connection_manager - ConnectionManager, keeps track of connected clients.
     * subscription_manager - SubscriptionManager, keeps track of subscribed clients.
     * notify_connections - boolean, set to True when using long-polling or streaming channels.
     * clean_freq - float - number of seconds to clean expired connections.
    """

    def __init__(self, service_mapper=None, connection_manager=None,
        subscription_manager=None, notify_connections=False, clean_freq=300):
        if service_mapper is None:
            from amfast.remoting import ServiceMapper
            service_mapper = ServiceMapper()
        self.service_mapper = service_mapper

        if connection_manager is None:
            connection_manager = cm.MemoryConnectionManager()
        self.connection_manager = connection_manager

        if subscription_manager is None:
            subscription_manager = sm.MemorySubscriptionManager()
        self.subscription_manager = subscription_manager

        self.notify_connections = notify_connections
        self.clean_freq = clean_freq
        self._lock = amfast.mutex_cls()
        self._channels = {}
        self.scheduleClean()

    def __iter__(self):
        # To make thread safe
        channel_vals = self._channels.values()
        return channel_vals.__iter__()

    def checkCredentials(self, user, password):
        """Determines if credentials are valid.

        arguments
        ==========
         * user - string, username.
         * password - string, password.

        Returns True if credentials are valid.
        Raises SecurityError if credentials are invalid.
        """
        raise SecurityError('Authentication not implemented.');

    def connect(self, channel, connection_id=None):
        """Add a client connection to this channel.

        arguments
        ==========
         * connection_id - string, Client connection id.

        Returns Connection
        """

        try:
            connection = self.connection_manager.getConnection(connection_id)
        except cm.NotConnectedError:
            pass
        else:
            raise ChannelError("Connection ID '%s' is already connected." % connection_id)

        return self.connection_manager.createConnection(channel, connection_id)

    def disconnect(self, connection):
        """Remove a client connection from this ChannelSet.

        arugments
        ==========
         * connection - Connection being disconnected.
        """

        self.subscription_manager.deleteConnection(connection)
        self.connection_manager.deleteConnection(connection)

    def scheduleClean(self):
        """Schedule connection cleaning procedure to run sometime in the future."""

        if amfast.use_dummy_threading is True:
            amfast.logger.warn('Connection cleaning was NOT scheduled.')
            return
        
        repeater = thread_pool.RepeatingThread(self.clean_freq, self.clean)
        repeater.start()

    def clean(self):
        """Clean out expired connections."""
        if amfast.log_debug is True:
            amfast.logger.debug("Cleaning channel.")

        current_time = time.time() * 1000
        if self.connection_manager.iterConnectionIds():
            for connection_id in self.connection_manager.iterConnectionIds():
                self.cleanConnection(connection_id, current_time)

        if hasattr(self.subscription_manager, 'deleteExpiredMessages'):
            # TODO: better interface for deleting expired messages.
            #
            # Work on this once async messaging support is dropped.
            #
            # Perhaps iterate through messages and delete expired
            # similar to how connection cleaning works??
            #
            # Currently, some subscriptions managers auto-delete
            # expired messages, while others require a method to be called.
            self.subscription_manager.deleteExpiredMessages(current_time)

    def cleanConnection(self, connection_id, current_time): 
        if amfast.log_debug is True: 
            amfast.logger.debug("Cleaning connection: %s" % connection_id) 
 
        try: 
            connection = self.connection_manager.getConnection(connection_id, False) 
        except cm.NotConnectedError:
            return
 
        if connection.last_active + connection.timeout < current_time: 
            channel = self.getChannel(connection.channel_name) 
            channel.disconnect(connection) 

    def publishObject(self, body, topic, sub_topic=None, headers=None, ttl=10000):
        """Create a message and publish it.

        arguments:
        ===========
        body - Any Python object.
        topic - string, the topic to publish to.
        sub_topic - string, the sub topic to publish to. Default = None
        headers - dict, headers to include with this message.
        ttl - int time to live in milliseconds. Default = 30000
        """

        if sub_topic is not None:
            if headers is None:
                headers = {}
            headers[messaging.AsyncMessage.SUBTOPIC_HEADER] = sub_topic

        current_time = time.time() * 1000

        msg = messaging.AsyncMessage(headers=headers, body=body,
            clientId=None, destination=topic, timestamp=current_time,
            timeToLive=ttl)
       
        self.publishMessage(msg)

    def publishMessage(self, msg):
        """Publish a pre-formed message.

        arguments:
        ===========
         * msg - AbstractMessage, the Flex message to publish.
        """

        self.subscription_manager.publishMessage(msg)

        if self.notify_connections is True:
            topic = msg.destination
            if hasattr(msg, 'headers') and \
                msg.headers is not None and \
                messaging.AsyncMessage.SUBTOPIC_HEADER in msg.headers:
                sub_topic = msg.headers[messaging.AsyncMessage.SUBTOPIC_HEADER]
            else:
                sub_topic = None

            self.notifyConnections(topic, sub_topic)

    def notifyConnections(self, topic, sub_topic):
        """Notify connections that a message has been published.

        arguments:
        ===========
         * topic - string, topic to notify
         * sub_topic - string, sub_topic to notify
        """
       
        thread_pool.GlobalThreadPool().addTask(self._notifyConnections, args=(topic, sub_topic)) 

    def _notifyConnections(self, topic, sub_topic):
        """Do the real work of notifyConnections."""

        for connection_id in self.subscription_manager.iterSubscribers(topic, sub_topic):
            try:
                connection = self.connection_manager.getConnection(connection_id, False)
            except cm.NotConnectedError:
                continue

            if connection.notify_func is not None:
                connection.notify_func()

    def mapChannel(self, channel):
        """Add a Channel to the ChannelSet

        arguments
        ==========
         * channel - Channel, the channel to add.
        """
        self._lock.acquire()
        try:
            self._channels[channel.name] = channel
            channel._channel_set = self
        finally:
            self._lock.release()

    def unMapChannel(self, channel):
        """Removes a Channel to the ChannelSet

        arguments
        ==========
         * channel - Channel, the channel to remove.
        """
        self._lock.acquire()
        try:
            if channel.name in self._channels:
                channel._channel_set = None
                del self._channels[channel.name]
        finally:
            self._lock.release()

    def getChannel(self, name):
        """Retrieves a Channel from the ChannelSet

        arguments
        ==========
         * name - string, the name of the Channel to retrieve.
        """
        try:
            return self._channels[name]
        except KeyError:
            raise ChannelError("Channel '%s' is not mapped." % name)



"""Provides an interface for performing remoting calls."""

import collections
import threading

import amfast
from amfast import AmFastError, class_def
from amfast.class_def.as_types import AsError

class RemotingError(AmFastError):
    """Remoting related errors."""
    pass

class Service(object):
    """A remoting service is a service that is exposed 
    by an amfast.remoting.channel.Channel to AMF clients. 

    attributes:
    ============
     * name - string, service name.
    """

    # Name of special service that handles packet header targets
    PACKET_HEADER_SERVICE = 'PACKET_HEADER_SERVICE'

    # Name of special service that handles command messages
    COMMAND_SERVICE = 'COMMAND_SERVICE'

    # Name of special service that handles targets without a service prefix
    DEFAULT_SERVICE = 'DEFAULT_SERVICE'

    SEPARATOR = '.' # Character used to separate service names and target names

    def __init__(self, name):
        self.name = name

        self._lock = threading.RLock()
        self._targets = {} # Keeps track of targets internally

    def __iter__(self):
        return self._targets.itervalues()

    def mapTarget(self, target):
        """Add a target to the service."""
        self._lock.acquire()
        try:
            self._targets[target.name] = target
        finally:
            self._lock.release()

    def unMapTarget(self, target):
        """Remove a target from the service."""
        self._lock.acquire()
        try:
            if target.name in self._targets:
                del self._targets[target.name]
        finally:
            self._lock.release()

    def getTarget(self, target_name):
        """Get a target from the service by name."""
        self._lock.acquire()
        try:
            target = self._targets.get(target_name, None)
        finally:
            self._lock.release()
        return target

class Target(object):
    """A remoting target can be invoked by an RPC message received from a client.

    attributes:
    ============
     * name - string, name of the target.
     * secure - boolean, True to require login.
    """

    @classmethod
    def mapObject(cls, service, obj, secure=False, map_private=False,\
        map_special=False):
        """Maps all instance methods of an object.

        arguments:
        ============
         * service - Service, the service to map to.
         * obj - object, the object instance to map methods from.
         * secure - boolean, True to require login. Default = False.
         * map_private - boolean, True to map methods starting with '_'.
                         Default = False
         * map_special - boolean, True to map methods starting with '__'.
                         Default = False
        """

        for attr in dir(obj):
            if map_special is False and attr.startswith('__'):
                continue
            if map_private is False and attr.startswith('_'):
                continue

            val = getattr(obj, attr, None)
            if isinstance(val, collections.Callable):
                service.mapTarget(cls(callable=val, name=attr, secure=secure))

    def __init__(self, name, secure=False):
        self.name = name
        self.secure = secure

    def _invokeStr(self, args):
        return "<targetInvocation target=\"%s\">%s</targetInvocation>" % \
            (self.name, args)

    def invoke(self, packet, msg, args):
        """Invoke a target.

        arguments
        ==========
         * packet - Packet, Packet that is invoking the target.
         * msg - Message, the message that is invoking this target.
             This value will be None if the target is being invoked by a package header.
         * args - list, list of arguments to pass to the callable.
        """
        raise RemotingError("'invoke' must be implemented on a sub-class.")

class CallableTarget(Target):
    """Calls an external callable with the passed arguments when invoked.

    attributes:
    ============
     * name - string, name of the target.
     * secure - boolean, True to require login.
     * callable - callable, a callable that can be invoked.
    """

    def __init__(self, callable, name, secure=False):
        Target.__init__(self, name, secure)
        self.callable = callable

    def invoke(self, packet, msg, args):
        """Calls self.callable and passes *args."""
        if amfast.log_debug:
            amfast.logger.debug(self._invokeStr(args))

        return self.callable(*args)

class ExtCallableTarget(CallableTarget):
    """Calls an external callable with the the same arguments that invoke receives. 

    attributes:
    ============
     * name - string, name of the target.
     * secure - boolean, True to require login.
     * callable - callable, a callable that can be invoked.
    """
    def invoke(self, packet, msg, args):
        if amfast.log_debug:
            amfast.logger.debug(self._invokeStr(args))

        return self.callable(packet, msg, *args)

class Header(object):
    """A remoting message header.

    attributes:
    ============
     * name - string, header name.
     * required - bool, True if header is required.
     * value - object, header value.
    """
    def __init__(self, name, required=False, value=None):
        self.name = name
        self.required = required
        self.value = value

    def __str__(self):
        return "<header name=\"%s\" required=\"%s\">%s</header>" % (self.name, self.required, self.value)

    def invoke(self, request):
        """Invoke an action on this header if one has been mapped."""
        target = request.channel.channel_set.service_mapper.\
            packet_header_service.getTarget(self.name)
        if target is not None:
            return target.invoke(request, None, (self.value,))
        return False

class Message(object):
    """A remoting message body.

    attributes:
    ============
     * target - Target, the target to be invoked.
     * response - string, message id.
     * body - object, message body.
    """

    SUCCESS_TARGET = '/onResult'
    FAILED_TARGET = '/onStatus'
    DEBUG_TARGET = '/onDebugEvents'

    def __init__(self, target=None, response=None, body=None):
        self.target = target
        self.response = response
        self.body = body

    def _isInvokable(self):
        """If True, the message's body can invoke itself."""
        if self.target == 'null':
            return True
        return False
    is_invokable = property(_isInvokable)

    def _isFlexMsg(self):
        """If True, the message's body is a Flex message."""
        return hasattr(self.body, 'FLEX_CLIENT_ID_HEADER')
    is_flex_msg = property(_isFlexMsg)

    def invoke(self, request):
        """Invoke an action on an RPC message and return a response message."""
        try:
            self.response_msg = self.acknowledge(request)
            if self.is_invokable:
                self.body[0].invoke(request, self)
            elif self.target is not None and self.target != '':
                self._invoke(request)
            else:
                raise RemotingError("Cannot invoke message: '%s'." % self)
        except Exception, exc:
            amfast.log_exc(exc)
            self.response_msg = self.fail(request, exc)

        return self.response_msg

    def _invoke(self, request):
        """Invoke an action on an AMF0 style RPC message."""
        qualified_name = self.target.split(Service.SEPARATOR)
        if len(qualified_name) < 2:
            target_name = self.target
            service_name = Service.DEFAULT_SERVICE
        else:
            target_name = qualified_name.pop()
            service_name = Service.SEPARATOR.join(qualified_name)

        target = request.channel.channel_set.service_mapper.getTarget(service_name, target_name)
        if target is None:
            raise RemotingError("Target '%s' not found." % self.target)

        if target.secure is True:
            # Make sure user is authenticated
            if not hasattr(request, '_authenticated'):
                raise RemotingError('Target requires authentication.');

        self.response_msg.body = target.invoke(request, self, self.body)

    def fail(self, request, exc):
        """Return an error response message."""
        response_target = self.response + self.FAILED_TARGET
        response_message = Message(target=response_target, response='')
        
        if self.is_invokable:
            error_val = self.body[0].fail(request, self, exc)
        else:
            error_val = AsError(exc=exc)

        response_message.body = error_val
        return response_message

    def convertFail(self, exc):
        """Convert a successful message into a failure."""
        self.target = self.target.replace(self.SUCCESS_TARGET, self.FAILED_TARGET)
        
        if self.is_flex_msg:
            self.body = self.body.convertFail(exc=exc)
        else:
            self.body = AsError(exc=exc)

    def acknowledge(self, request):
        """Return a successful response message to acknowledge an RPC message."""
        response_target = self.response + self.SUCCESS_TARGET
        response_message = Message(target=response_target, response='')
        
        if self.is_invokable:
            response_message.body = self.body[0].acknowledge(request, self)

        return response_message

    def __str__(self):
        return "<message> <target>%s</target> <response>%s</response> <body>%s</body></message>" % (self.target, self.response, self.body)

class Packet(object):
    """An AMF NetConnection packet that can be passed from client->server or server->client.

    attributes:
    ============
     * client_type - string, the type of client connected to the server.
     * headers - dict, keys = header names, values = Header objects.
     * messages - list, a list of messages that belong to the packet.
    """

    FLASH_8 = 0x00
    FLASH_COM = 0x01
    FLASH_9 = 0x03

    def __init__(self, client_type=None, headers=None, messages=None):
        if client_type is None:
            client_type = self.FLASH_8
        self.client_type = client_type 

        if headers is None:
            headers = []
        self.headers = headers

        if messages is None:
            messages = []
        self.messages = messages

    def _getAmf3(self):
        if self.client_type == self.FLASH_9:
            return True
        return False
    is_amf3 = property(_getAmf3)

    def invoke(self):
        """Process an RPC packet and return a response packet."""
        if amfast.log_debug:
            amfast.logger.debug("<requestPacket>%s</requestPacket>" % self)

        self.response = self.acknowledge()
        try:
            # Invoke any headers
            for header in self.headers:
                header.invoke(self)

            # Invoke any messages
            for message in self.messages:
                self.response.messages.append(message.invoke(self))
        except Exception, exc:
            # Fail all messages
            amfast.log_exc(exc)
            self.response = self.fail(exc)

        if amfast.log_debug:
            amfast.logger.debug("<responsePacket>%s</responsePacket>" % self.response)

        return self.response

    def fail(self, exc):
        """Return a response Packet with all messages failed."""
        response = self.acknowledge()

        for message in self.messages:
            response.messages.append(message.fail(self, exc))
        return response

    def acknowledge(self):
        """Create a response to this packet."""
        response = Packet()
        response.client_type = self.client_type
        return response

    def __str__(self):
        header_msg = "\n  ".join(["%s" % header for header in self.headers])

        message_msg = "\n  ".join(["%s" % message for message in self.messages])

        return """
<Packet>
 <headers>
  %s
 </headers>

 <messages>
  %s
 </messages>

 <attributes>
  <attr name="client_type">%s</attr>
 </attributes>
</Packet>
""" % (header_msg, message_msg, self.client_type)

class ServiceMapper(object):
    """Maps service to service name.

    attributes
    ===========
     * packet_header_service - Service, a special service for AMF packet headers.
     * message_header_service - Service, a special service for AMF message headers.
     * command_service - Service, a special service for Flex CommandMessages.
     * default_service - Service, a special service for targets that don't have service specifiers.
    
    When an AMF packet or message is processed, the object checks
    the header_service Service for a target where target.name == header.name
    for each header. If a target is found, the target will be invoked
    before any packet.messages are invoked.

    An example of how to use this functionality is adding a target named 'Credentials'
    to packet_header_service that checks credentials stored in the 'Credentials' header before
    invoking any messages in the packet.
    """

    def __init__(self):
        self._services = {} # used internally to keep track of Service objects.
        self._mapBuiltIns()
        self._lock = threading.RLock()

    def __iter__(self):
        return self._services.itervalues()

    def _mapBuiltIns(self):
        """Map default Targets required for Authentication and FlexMessaging.

        Users can override the defaults, by remapping their own targets.
        """ 
        import amfast.remoting.flex_messages as messaging
        import targets

        # Map built in targets
        self.packet_header_service = Service(Service.PACKET_HEADER_SERVICE)
        self.mapService(self.packet_header_service)
        self.command_service = Service(Service.COMMAND_SERVICE)
        self.mapService(self.command_service)
        self.default_service = Service(Service.DEFAULT_SERVICE)
        self.mapService(self.default_service)

        # NetConnection authentication
        self.packet_header_service.mapTarget(ExtCallableTarget(targets.nc_auth,
            'Credentials'))

        # CommandMessages
        self.command_service.mapTarget(ExtCallableTarget(targets.client_ping,
            messaging.CommandMessage.CLIENT_PING_OPERATION))
        self.command_service.mapTarget(ExtCallableTarget(targets.login_operation,
            messaging.CommandMessage.LOGIN_OPERATION))
        self.command_service.mapTarget(ExtCallableTarget(targets.logout_operation,
            messaging.CommandMessage.LOGOUT_OPERATION))
        self.command_service.mapTarget(ExtCallableTarget(targets.poll_operation,
            messaging.CommandMessage.POLL_OPERATION))
        self.command_service.mapTarget(ExtCallableTarget(targets.subscribe_operation,
            messaging.CommandMessage.SUBSCRIBE_OPERATION))
        self.command_service.mapTarget(ExtCallableTarget(targets.unsubscribe_operation,
            messaging.CommandMessage.UNSUBSCRIBE_OPERATION))
        self.command_service.mapTarget(ExtCallableTarget(targets.disconnect_operation,
            messaging.CommandMessage.DISCONNECT_OPERATION))

    def mapService(self, service):
        """Maps a service

        arguments
        ==========
         * service - Service, the service to map.
        """
        self._services[service.name] = service

    def unMapService(self, service):
        """Un-maps a service

        arguments
        ==========
         * service - Service, the service to un-map.
        """
        self._lock.acquire()
        try:
            if service.name in self._services:
                del self._services[service.name]
        finally:
            self._lock.release()

    def getTarget(self, service_name, target_name):
        """Get a Target

        Returns None in Target is not found.

        arguments
        ==========
         * service_name - string, the service name.
         * target_name - string, the target name.
        """
        self._lock.acquire()
        try:
            service = self.getService(service_name)
            if service is None:
                target = None
            else:
                target = service.getTarget(target_name)
        finally:
            self._lock.release()

        return target

    def getService(self, service_name):
        """Get a Service

        Returns None in Service is not found.

        arguments
        ==========
         * service_name - string, the service name.
        """
        return self._services.get(service_name, None)



"""Classes to represent Actionscript types that have no-equivalent in Python."""

import amfast

class AsByteArray(object):
    """An Actionscript ByteArray.

    attributes
    ===========
    bytes - string, bytes.
    """
    
    AS_BYTE_ARRAY = True

    def __init__(self, bytes):
        self.bytes = bytes

class AsProxy(object):
    """A proxy object.

    Forces an object to be encoded as a proxy
    even if use_collection or use_proxies if False.

    attributes
    ===========
    source - object, the proxy source.
    """

    AS_PROXY = True

    def __init__(self, source=None):
        self.source = source

class AsNoProxy(object):
    """A no-proxy object.

    Forces an object to be encoded without a proxy
    even if use_collection or use_proxies is True.

    attributes
    ===========
    source - object, the no-proxy source.
    """

    AS_NO_PROXY = True

    def __init__(self, source=None):
        self.source = source

class AsError(amfast.AmFastError):
    """Equivalent to: 'Error' in AS."""

    APPLICATION_ERROR = 5000

    def __init__(self, message='', exc=None):
        self.errorID = self.APPLICATION_ERROR
        if exc is not None:
            self.name = exc.__class__.__name__
            self.message = "%s" % exc
        else:
            self.name = ''
            self.message = message

        amfast.AmFastError.__init__(self, self.message)
amfast.class_def.assign_attrs(AsError, 'Error', ('errorId', 'name', 'message'), amf3=False)



"""Built in Target functions."""
import base64

from amfast.class_def.as_types import AsNoProxy
from amfast.remoting.flex_messages import CommandMessage
from amfast.remoting.channel import ChannelError, SecurityError
from amfast.remoting.endpoint import AmfEndpoint

# ---- NetConnection Operations --- #

def nc_auth(packet, msg, credentials):
    """NetConnection style authentication."""
    packet.channel.channel_set.checkCredentials(
        credentials['userid'], credentials['password'])

    # Flag indicating packet was 
    # authenticated properly.
    packet._authenticated = True

# --- Flex CommandMessage Operations --- #

def client_ping(packet, msg, *args):
    """Respond to a ping request and connect to the Channel."""

    response = msg.response_msg.body
    if (not hasattr(response, 'headers')) or response.headers is None:
        response.headers = {}

    command = msg.body[0]
    response.headers[command.FLEX_CLIENT_ID_HEADER] = command.connection.id

def login_operation(packet, msg, raw_creds):
    """RemoteObject style authentication."""

    cred_str = base64.decodestring(raw_creds)

    command = msg.body[0]
    if hasattr(command, 'headers') and \
        command.CREDENTIALS_CHARSET_HEADER in command.headers:
        # Convert encoded string
        cred_str = unicode(cred_str, command.headers[command.CREDENTIALS_CHARSET_HEADER])

    creds = cred_str.split(':')
    
    channel_set = packet.channel.channel_set
    channel_set.checkCredentials(creds[0], creds[1])
    command.connection.authenticate(creds[0])

def logout_operation(packet, msg, *args):
    """RemoteObject style de-authentication."""

    command = msg.body[0]
    command.connection.unAuthenticate()

def disconnect_operation(packet, msg, *args):
    """Respond to a disconnect operation. Disconnects current Connection."""

    command = msg.body[0]
    packet.channel.disconnect(command.connection)

    response = msg.response_msg.body
    if hasattr(response, 'headers') and response.FLEX_CLIENT_ID_HEADER in response.headers:
        del response.headers[response.FLEX_CLIENT_ID_HEADER]

def subscribe_operation(packet, msg, *args):
    """Respond to a subscribe operation."""
    command = msg.body[0]

    channel_set = packet.channel.channel_set
    if channel_set.subscription_manager.secure is True:
        if command.connection.authenticated is False:
            raise SecurityError("Operation requires authentication.")

    headers = command.headers
    channel_set.subscription_manager.subscribe(command.connection.id,
        command.clientId, command.destination,
        headers.get(command.SUBTOPIC_HEADER, None),
        headers.get(command.SELECTOR_HEADER, None))

def unsubscribe_operation(packet, msg, *args):
    """Respond to an unsubscribe operation."""
    command = msg.body[0]
    packet.channel.channel_set.subscription_manager.unSubscribe(
        command.connection.id, command.clientId, command.destination,
        command.headers.get(command.SUBTOPIC_HEADER, None))

def poll_operation(packet, msg, *args):
    """Respond to a poll operation. Returns queued messages."""
    command = msg.body[0]
    connection = command.connection
    channel = packet.channel

    msgs = channel.channel_set.subscription_manager.pollConnection(connection)
    if len(msgs) < 1 and channel.wait_interval != 0:
        # Long polling channel, don't return response
        # until a message is available.
        msgs = channel.waitForMessage(packet, msg, connection)

    if isinstance(channel.endpoint, AmfEndpoint):
        # Make sure messages are not encoded as an ArrayCollection
        return AsNoProxy(msgs)
    else:
        return msgs



"""Provides an interface for determining how Python objects are serialized and de-serialized."""
import threading

from amfast import AmFastError

class ClassDefError(AmFastError):
    """ClassDef related errors."""
    pass

class ClassDef(object):
    """Defines how objects of a given class are serialized and de-serialized.
    This class can be sub-classed to provide custom serialization.

    attributes
    ===========
     * class_ - class, the class object mapped to this definition
     * alias - string, the AMF alias name of the mapped class
     * static_attrs - tuple or list, a tuple of static attribute names,
          all values must be strings or unicode.
     * amf3 - bool, if True, this object will be encoded in AMF3.
     * encode_types - dict, keys = attribute names, values = callables.
         Callables must accept a single parameter
         (the object being encoded) and return a new object.
     * decode_types - dict, keys = attribute names, values = callables.
         Callables must accept a single parameter
         (the object being decoded) and return a new object.

    """

    CLASS_DEF = True

    def __init__(self, class_, alias=None, static_attrs=None,
        amf3=None, encode_types=None, decode_types=None, _built_in=False):
        """arguments
        =============
         * class_ - class, the class being mapped.
         * alias - string, specifies the amf class alias. Default = module.class
         * static_attrs - tuple or list, a tuple of static attribute strings. Default = empty tuple
         * amf3 - bool, if True, this object will be encoded in AMF3. Default = True
         * encode_types - dict, keys = attribute names, values = callables. Default = None
         * decode_types - dict, keys = attribute names, values = callables. Default = None
        """
        self.class_ = class_
        self._built_in = _built_in

        if alias is None:
            if hasattr(class_, ALIAS):
                alias = getattr(class_, ALIAS)
            else:
                alias = '.'.join((class_.__module__, class_.__name__))
        self.alias = alias

        if static_attrs is None:
            if hasattr(class_, STATIC_ATTRS):
                static_attrs = self.static_attrs = getattr(class_, STATIC_ATTRS)
            else: 
                static_attrs = () 
        self.static_attrs = static_attrs

        if amf3 is None:
            if hasattr(class_, AMF3):
                amf3 = getattr(class_, AMF3)
            else:
                amf3 = True
        self.amf3 = amf3

        self.encode_types = encode_types
        self.decode_types = decode_types

    def getStaticAttrVals(self, obj):
        """Returns a list of values of attributes defined in self.static_attrs

        If this method is overridden to provide custom behavior, please note:
        Returned values MUST BE IN THE SAME ORDER AS THEY APPEAR IN self.static_attrs.

        arguments
        ==========
         * obj - object, the object to get attribute values from.
        """
        return [getattr(obj, attr, None) for attr in self.static_attrs]

    def getInstance(self):
        """Returns an instance of the mapped class to be used 
        when an object of this type is deserialized.
        """
        return self.class_.__new__(self.class_)

    def applyAttrVals(self, obj, vals):
        """Set decoded attribute values on the object.

        arguments
        ==========
         * obj - object, the object to set the attribute values on.
         * vals - dict, keys == attribute name, values == attribute values.
        """
        [setattr(obj, key, val) for key, val in vals.iteritems()]

class DynamicClassDef(ClassDef):
    """A ClassDef with dynamic attributes."""

    DYNAMIC_CLASS_DEF = True

    def __init__(self, class_, alias=None, static_attrs=None, amf3=True,
        encode_types=None, decode_types=None, include_private=None, _built_in=False):
        ClassDef.__init__(self, class_, alias, static_attrs, amf3,
            encode_types, decode_types, _built_in)

        self.include_private = include_private

    def getDynamicAttrVals(self, obj, include_private=False):
        """Returns a dict where keys are attribute names and values are attribute values.

        arguments
        ==========
        obj - object, the object to get attributes for.
        include_private - bool, if False do not include attributes with
            names starting with '_'. Default = False.
        """
        if self.include_private is None:
            ip = include_private
        else:
            ip = self.include_private

        return get_dynamic_attr_vals(obj, self.static_attrs, ip);

class ExternClassDef(ClassDef):
    """A ClassDef where the byte string encoding/decoding is customized.
   
    The Actionscript version of the class must implement IExternalizeable.
    """
    
    EXTERNALIZABLE_CLASS_DEF = True
    
    def __init__(self, class_, alias=None, static_attrs=None, _built_in=False):
        ClassDef.__init__(self, class_, alias, static_attrs, amf3=True,
            _built_in=_built_in)
 
    def writeExternal(self, obj, context):
        """
        This method must be overridden in a sub-class.

        arguments
        ==========
         * obj - object, The object that is being encoded.
         * context - amfast.decoder.EncoderContext, holds encoding related properties.
        """
        raise ClassDefError("This method must be implemented by a sub-class.")

    def readExternal(self, obj, context):
        """
        This method must be overridden in a sub-class.

        arguments
        ==========
         * obj - object, The object that the byte string is being applied to.
         * context - amfast.decoder.DecoderContext, holds decoding related properties.
        """
        raise ClassDefError("This method must be implemented by a sub-class.")

class _ProxyClassDef(ExternClassDef):
    """A special class used internally to encode/decode Proxied objects."""

    PROXY_CLASS_DEF = True
    PROXY_ALIAS = 'proxy'

    class _ProxyObject(object):
        """Empty class used for mapping."""
        pass

    def __init__(self):
        ExternClassDef.__init__(self, self._ProxyObject, self.PROXY_ALIAS,
            None, _built_in=True)

class _ArrayCollectionClassDef(_ProxyClassDef):
    """A special ClassDef used internally to encode/decode an ArrayCollection."""

    ARRAY_COLLECTION_CLASS_DEF = True
    PROXY_ALIAS = 'flex.messaging.io.ArrayCollection'

    def __init__(self):
        _ProxyClassDef.__init__(self)

class _ObjectProxyClassDef(_ProxyClassDef):
    """A special ClassDef used internally to encode/decode an ObjectProxy."""

    OBJECT_PROXY_CLASS_DEF = True
    PROXY_ALIAS = 'flex.messaging.io.ObjectProxy'

    def __init__(self):
        _ProxyClassDef.__init__(self)

class ClassDefMapper(object):
    """Map classes to ClassDefs, retrieve class_defs by class or alias name."""
    def __init__(self):
        """
        arguments
        ==========
        * class_def_attr - string, an attribute with this name will be added
            mapped classes. Default = '_amf_alias'
        """

        self._lock = threading.RLock()
        self._mapped_classes = {}
        self._mapped_aliases = {}
        self._mapBuiltIns()

    def __iter__(self):
        return self._mapped_aliases.itervalues()

    def _mapBuiltIns(self):
        """Map built-in ClassDefs for default behavior."""
        from as_types import AsError
        from amfast.remoting import flex_messages as messaging

        # Proxy objects
        self.mapClass(_ArrayCollectionClassDef())
        self.mapClass(_ObjectProxyClassDef())

        # Exceptions
        self.mapClass(ClassDef(AsError, _built_in=True))
        self.mapClass(ClassDef(messaging.FaultError, _built_in=True))

        # Flex remoting messages
        self.mapClass(ClassDef(messaging.RemotingMessage, _built_in=True))
        self.mapClass(messaging.AsyncSmallMsgDef(messaging.AsyncMessage,
            alias="DSA", _built_in=True))
        self.mapClass(ClassDef(messaging.AsyncMessage, _built_in=True))
        self.mapClass(messaging.CommandSmallMsgDef(messaging.CommandMessage,
            alias="DSC", _built_in=True))
        self.mapClass(ClassDef(messaging.CommandMessage, _built_in=True))
        self.mapClass(ClassDef(messaging.AcknowledgeMessage, _built_in=True))
        self.mapClass(ClassDef(messaging.ErrorMessage, _built_in=True))

    def mapClass(self, class_def):
        """Map a class_def implementation, so that it can be retrieved based on class attributes.

        arguments
        ==========
         * class_def - ClassDef, ClassDef being mapped.
        """
        if not hasattr(class_def, 'CLASS_DEF'):
            raise ClassDefError("class_def argument must be a ClassDef object.")

        self._lock.acquire()
        try:
            self._mapped_classes[class_def.class_] = class_def
            self._mapped_aliases[class_def.alias] = class_def
        finally:
            self._lock.release()

    def getClassDefByClass(self, class_):
        """Get a ClassDef.

        Returns None if ClassDef is not found.

        arguments
        ==========
         * class_ - class, the class to find a ClassDef for.
        """
        return self._mapped_classes.get(class_, None)

    def getClassDefByAlias(self, alias):
        """Get a ClassDef.

        Returns None in not ClassDef is found.

        arguments
        ==========
         * alias - string, the alias to find a ClassDef for.
        """
        return self._mapped_aliases.get(alias, None)

    def unmapClass(self, class_):
        """Unmap a class definition.

        arguments
        ==========
         * class_ - class, the class to remove a ClassDef for.
        """
        self._lock.acquire()
        try:
            for alias, klass in self._mapped_aliases.iteritems():
                if class_ == klass:
                    del self._mapped_aliases[alias]

            class_id = id(class_)
            if class_id in self._mapped_classes:
                del self._mapped_classes[class_id]
        finally:
            self._lock.release()

# ---- module attributes ---- #

def get_dynamic_attr_vals(obj, ignore_attrs=None, include_private=False):
    """Returns a dict of attribute values to encode.

    keys = attribute names, values = attribute values.

    argmuents
    ==========
     * obj - object, object to get dynamic attribute values from.
     * ignore_attrs - list or tuple of attributes to ignore. Default = empty tuple.
     * include_private - bool, if False do not include attributes that start with '_'.
          Default = False.
    """ 
    vals = {}

    if hasattr(obj, '__dict__'):
        for attr, val in obj.__dict__.iteritems():
            if ignore_attrs is not None:
                if attr in ignore_attrs:
                    continue

            if (include_private is False) and (attr.startswith('_')):
                continue

            vals[attr] = val

    return vals

# These properties can be set on a class
# to map attributes within the class.
ALIAS = '_AMFAST_ALIAS'
STATIC_ATTRS = '_AMFAST_STATIC_ATTRS'
AMF3 = '_AMFAST_AMF3'

def assign_attrs(class_, alias=None, static_attrs=None, amf3=None):
    """
    Use to map ClassDef attributes to a class. Useful if you want to keep
    ClassDef configuration with the class being mapped, instead of at 
    the point where the ClassDef is created.

    If you assign ClassDef attributes with this method, you can
    call ClassDef(class_) to create a ClassDef, and the assigned
    attributes will be applied to the new ClassDef.

    Arguments provided to the ClassDef() will override attributes
    that were assigned with this function.

    arguments
    ==========
     * class_ - class, the class to assign attributes to.
     * alias - string, the amf alias name of the mapped class
     * static_attrs - tuple, a tuple of static attribute names, all values must be strings or unicode
     * amf3 - bool, if True, this object will be encoded in AMF3.
    """
    if alias is not None:
        setattr(class_, ALIAS, alias)

    if static_attrs is not None:
        setattr(class_, STATIC_ATTRS, static_attrs)

    if amf3 is not None:
        setattr(class_, AMF3, amf3)



"""Use PyAmf for encoding and decoding."""

import logging

import pyamf
from pyamf import util as pyamf_util
import pyamf.remoting

import amfast
from endpoint import Endpoint
import pyamf_converter as pc

class PyAmfEndpoint(Endpoint):
    """An Endpoint that can encode/decode AMF packets with PyAmf.

    How to configure custom class mapping:

# When using the PyAmfEndpoint,
# custom type mapping can be configured
# either through AmFast, or through PyAmf.

# Configure type mapping with AmFast
class_mapper = ClassDefMapper()

#... map classes ...#

# Use pyamf_converter to automatically map classes
# from a ClassDefMapper with PyAmf.
import amfast.remoting.pyamf_converter as pyamf_converter
pyamf_converter.register_class_mapper(class_mapper)

# Configure type mapping directly with PyAmf.
# Use the standard PyAmf way of mapping classes.
pyamf.register_class(klass, 'alias', ....)
    """

    def decodePacket(self, raw_packet, *args, **kwargs):
        if amfast.log_raw:
            self.logRaw('rawDecodePacket', raw_packet)

        context = pyamf.get_context(pyamf.AMF0)
        pyamf_packet = pyamf.remoting.decode(raw_packet, context)
        packet = pc.packet_to_amfast(pyamf_packet)

        if amfast.log_debug:
            amfast.logger.debug("<decodedPyAmfPacket>%s</decodedPyAmfPacket>" % pyamf_packet)

        return packet

    def encodePacket(self, packet):
        pyamf_packet = pc.packet_to_pyamf(packet)
        if amfast.log_debug:
            amfast.logger.debug("<encodedPyAmfPacket>%s</encodedPyAmfPacket>" % pyamf_packet)

        context = pyamf.get_context(pyamf.AMF0)
        stream = pyamf.remoting.encode(pyamf_packet, context)
        raw_packet = stream.getvalue()

        if amfast.log_raw:
            self.logRaw('rawEncodePacket', raw_packet)

        return raw_packet

    def decode(self, raw_obj, amf3=False):
        if amf3 is True:
            amf_type = pyamf.AMF3
        else:
            amf_type = pyamf.AMF0

        context = pyamf.get_context(amf_type)
        decoder = pyamf.get_decoder(amf_type, raw_obj, context=context)
        obj = decoder.readElement()

        if amfast.log_raw:
            self.logRaw('rawDecodeObject', raw_obj)

        return obj

    def encode(self, obj, amf3=False):
        if amf3 is True:
            amf_type = pyamf.AMF3
        else:
            amf_type = pyamf.AMF0

        stream = pyamf_util.BufferedByteStream()

        context = pyamf.get_context(amf_type)
        encoder = pyamf.get_encoder(amf_type, stream, context=context)
        encoder.writeElement(obj)

        raw_obj = stream.getvalue()

        if amfast.log_raw:
            self.logRaw('rawEncodeObject', raw_obj)

        return raw_obj



import unittest
import time
import random

from amfast.remoting.channel import ChannelSet, Channel, HttpChannel, ChannelError
from amfast.remoting.connection_manager import NotConnectedError

class MessagingTestCase(unittest.TestCase):

    CHANNEL_NAME = 'channel'
    HTTP_CHANNEL_NAME = 'http_channel'
    TOPIC = 'topic'

    # Use the same ChannelSet for everything
    channel_set = ChannelSet()
    channel_set.mapChannel(Channel(CHANNEL_NAME))
    channel_set.mapChannel(HttpChannel(HTTP_CHANNEL_NAME))

    def setUp(self):
        self.channel_set.connection_manager.reset()
        self.channel_set.subscription_manager.reset()
        self.flex_client_id = 'my_client_id'

    def tearDown(self):
        pass

    def connect(self, channel_name):
        channel = self.channel_set.getChannel(channel_name)
        return channel.connect()

    def testConnect(self):
        connection = self.connect(self.CHANNEL_NAME)

        channel = self.channel_set.getChannel(self.CHANNEL_NAME)
        self.assertTrue(channel.channel_set.connection_manager.getConnectionCount(channel.name) > 0)

        channel = self.channel_set.getChannel(self.HTTP_CHANNEL_NAME)
        self.assertRaises(ChannelError, channel.connect, connection.id)

    def testDisconnect(self):
        connection = self.connect(self.CHANNEL_NAME)
        channel = self.channel_set.getChannel(self.CHANNEL_NAME)

        connection_count = channel.channel_set.connection_manager.getConnectionCount(channel.name)
        id = connection.id
        channel.disconnect(connection)
         
        self.assertEquals(connection_count - 1, channel.channel_set.connection_manager.getConnectionCount(channel.name))
        self.assertRaises(NotConnectedError, self.channel_set.connection_manager.getConnection, id)

    def testSubscribe(self):
        connection = self.connect(self.CHANNEL_NAME)
        self.channel_set.subscription_manager.subscribe(connection, self.flex_client_id, self.TOPIC)

    def testUnSubscribe(self):
        connection = self.connect(self.CHANNEL_NAME)
        self.channel_set.subscription_manager.subscribe(connection, self.flex_client_id, self.TOPIC)
        
        self.channel_set.subscription_manager.unSubscribe(connection, self.flex_client_id, self.TOPIC)

    def testPublish(self):
        connection = self.connect(self.CHANNEL_NAME)
        self.channel_set.subscription_manager.subscribe(connection.id, self.flex_client_id, self.TOPIC)

        self.channel_set.publishObject('test', self.TOPIC)
        msgs = self.channel_set.subscription_manager.pollConnection(connection)
        self.assertEquals(1, len(msgs))

    def testPoll(self):
        connection = self.connect(self.CHANNEL_NAME)
        self.channel_set.subscription_manager.subscribe(connection.id, self.flex_client_id, self.TOPIC)

        # Make sure messages are polled
        self.channel_set.publishObject('test', self.TOPIC)
        msgs = self.channel_set.subscription_manager.pollConnection(connection)
        self.assertEquals(1, len(msgs))

        # Make sure polled messages were deleted
        msgs = self.channel_set.subscription_manager.pollConnection(connection)
        self.assertEquals(0, len(msgs))

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(MessagingTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



import time
import uuid

import amfast
from connection import Connection, ConnectionError
import flex_messages as messaging

class NotConnectedError(ConnectionError):
    pass

class SessionAttrError(ConnectionError):
    pass

class ConnectionManager(object):
    """Keeps track of all current connections.

    This is an abstract base class and should be 
    implemented by a sub-class.

    """

    def __init__(self, connection_class=Connection, connection_params=None):
        self.connection_class = connection_class
        if connection_params is None:
            connection_params = {}
        self.connection_params = connection_params

    def generateId(self):
        """Generates a unique ID for a connection."""
        return str(uuid.uuid4())

    def getConnection(self, connection_id, touch=True):
        """Retrieve an existing connection.

        arugments
        ==========
         * connection_id - string, id of client to get connection for.
             connection_id should be unique for each Flash client (flex_client_id).
         * touch - boolean, If True set 'last_accessed' to now.

        raises
        =======
         * NotConnectedError if connection doesn't exist.
        """
        if connection_id is None:
            raise NotConnectedError("Blank connection_id is not connected.")

        connection = self.loadConnection(connection_id)

        if touch is True:
            self.touchConnection(connection)
        else:
            # Check for timeout
            current_time = time.time() * 1000
            if connection.last_active < (current_time - connection.timeout):
                connection.delete()
                raise NotConnectedError("Connection '%s' is not connected." % connection_id)

        return connection

    def createConnection(self, channel, connection_id=None):
        """Returns a new connection object."""
        if connection_id is None:
            connection_id = self.generateId()

        connection = self.connection_class(self, channel.name, connection_id, **self.connection_params)
        self.initConnection(connection, channel)
        return connection

    def deleteConnection(self, connection):
        """Deletes a connection object."""
        if connection.notify_func is not None:
            # Call notify function,
            # which should check for
            # a disconnection.
            connection.notify_func()
            connection.unSetNotifyFunc()

        connection.disconnect()

    def setNotifyFunc(self, connection, func):
        raise ConnectionError('Not implemented')

    def unSetNotifyFunc(self, connection):
        raise ConnectionError('Not implemented')

class MemoryConnectionManager(ConnectionManager):
    """Manages connections in memory."""

    def __init__(self, connection_class=Connection, connection_params=None):
        ConnectionManager.__init__(self, connection_class=connection_class,
            connection_params=connection_params)

        self._lock = amfast.mutex_cls()
        self.reset()

    def reset(self):
        self._connections = {}
        self._channels = {}

    def getConnectionCount(self, channel_name):
        try:
            return self._channels[channel_name]
        except KeyError:
            return 0

    def loadConnection(self, connection_id):
        connection = self._connections.get(connection_id, None)
        if connection is None:
            raise NotConnectedError("Connection '%s' is not connected." % connection_id)

        return connection

    def initConnection(self, connection, channel):
        connection._session = {}
        connection._connected = True
        connection._last_active = time.time() * 1000
        connection._last_polled = 0.0
        connection._authenticated = False
        connection._flex_user = None

        self._connections[connection.id] = connection

        self._lock.acquire()
        try:
            try:
                self._channels[channel.name] += 1
            except KeyError:
                self._channels[channel.name] = 1
        finally:
            self._lock.release()

    def iterConnectionIds(self):
        return self._connections.keys().__iter__()

    # --- proxies for connection properties --- #

    def getConnected(self, connection):
        return connection._connected

    def getLastActive(self, connection):
        return connection._last_active

    def getLastPolled(self, connection):
        return connection._last_polled

    def getAuthenticated(self, connection):
        return connection._authenticated

    def getFlexUser(self, connection):
        return connection._flex_user

    def getNotifyFunc(self, connection):
        if not hasattr(connection, '_notify_func_id'):
            return None
        else:
            return connection._getNotifyFuncById(connection._notify_func_id)

    # --- proxies for connection methods --- #

    def connectConnection(self, connection):
        connection._connected = True

    def disconnectConnection(self, connection):
        connection._connected = False

    def deleteConnection(self, connection):
        self._lock.acquire()
        try:
            if connection.id in self._connections:
                del self._connections[connection.id]

            if connection.channel_name in self._channels:
                self._channels[connection.channel_name] -= 1
        finally:
            self._lock.release()

        ConnectionManager.deleteConnection(self, connection)

    def touchConnection(self, connection):
        connection._last_active = time.time() * 1000

    def touchPolled(self, connection):
        connection._last_polled = time.time() * 1000

    def authenticateConnection(self, connection, user):
        connection._authenticated = True
        connection._flex_user = user

    def unAuthenticateConnection(self, connection):
        connection._authenticated = False
        connection._flex_user = None

    def setNotifyFunc(self, connection, func):
        connection._notify_func_id = connection._setNotifyFunc(func)

    def unSetNotifyFunc(self, connection):
        if hasattr(connection, '_notify_func_id'):
            connection._delNotifyFunc(connection._notify_func_id)
            del connection._notify_func_id

    def getConnectionSessionAttr(self, connection, name):
        try:
            return connection._session[name]
        except KeyError:
            raise SessionAttrError("Attribute '%s' not found." % name)

    def setConnectionSessionAttr(self, connection, name, val):
        connection._session[name] = val

    def delConnectionSessionAttr(self, connection, name):
        try:
            del connection._session[name]
        except KeyError:
            pass



import time

from twisted.internet import defer, task, reactor
from twisted.web import server
from twisted.web.resource import Resource

import amfast
from amfast.class_def.as_types import AsNoProxy
from channel import ChannelSet, HttpChannel
from endpoint import AmfEndpoint
import flex_messages as messaging
import connection_manager as cm

class TwistedChannelSet(ChannelSet):
    """A ChannelSet optimized for use with Twisted."""

    def render_POST(self, request):
        channel_name = request.path[1:]
        channel = self.getChannel(channel_name)
        return channel.render_POST(request)

    def scheduleClean(self):
        cleaner = task.LoopingCall(self.clean)
        cleaner.start(self.clean_freq, False)

    def notifyConnections(self, topic, sub_topic):
        reactor.callLater(0, self._notifyConnections, topic, sub_topic)

    def _notifyConnections(self, topic, sub_topic):
        iter = self.subscription_manager.iterSubscribers(topic, sub_topic)
        def _notify():
            try:
                connection_id = iter.next()
                try:
                    connection = self.connection_manager.getConnection(connection_id, False)
                except cm.NotConnectedError:
                    pass
                else:
                    if connection.notify_func is not None:
                        reactor.callLater(0, connection.notify_func)
                reactor.callLater(0, _notify)
            except StopIteration:
                pass
        reactor.callLater(0, _notify)

    def clean(self):
        if amfast.log_debug is True:
            amfast.logger.debug("Cleaning channel.")

        current_time = time.time()
        iter = self.connection_manager.iterConnectionIds()

        def _clean():
            try:
                connection_id = iter.next()
                reactor.callLater(0, self.cleanConnection, connection_id, current_time)
                reactor.callLater(0, _clean)
            except StopIteration:
                pass
        reactor.callLater(0, _clean)

class TwistedChannel(Resource, HttpChannel):
    """An AMF messaging channel that can be used with Twisted Web."""

    # This attribute is added to packets
    # that are waiting for a long-poll
    # to receive a message.
    MSG_NOT_COMPLETE = '_msg_not_complete'

    # This attribute is added to store
    # Twisted's request var on the packet,
    # so that it can be available to targets.
    TWISTED_REQUEST = '_twisted_request'

    def __init__(self, *args, **kwargs):

        Resource.__init__(self)
        HttpChannel.__init__(self, *args, **kwargs)

    def render_POST(self, request):
        """Process an incoming AMF packet."""
        if request.content:
            d = defer.Deferred()
            d.addCallbacks(request.content.read, self.fail, errbackArgs=(request,))
            d.addCallbacks(self.decode, self.fail, callbackArgs=(request,), errbackArgs=(request,))
            d.addCallbacks(self.invoke, self.fail, errbackArgs=(request,))
            d.addCallbacks(self.checkComplete, self.fail, callbackArgs=(request,), errbackArgs=(request,))
            d.callback(int(request.getHeader('Content-Length')))

            return server.NOT_DONE_YET

    def decode(self, raw_request, request):
        """Overridden to add Twisted's request object onto the packet."""
        decoded = HttpChannel.decode(self, raw_request)
        setattr(decoded, self.TWISTED_REQUEST, request)
        return decoded

    def getDeferred(self, msg):
        """Returns a Deferred object if a message contains a Deferred in its body,
        or False if message is not deferred.
        """
        if msg.is_flex_msg:
            body = msg.body.body
            if isinstance(body, defer.Deferred):
                return body
        else:
            body = msg.body
            if isinstance(body, defer.Deferred):
                return body
        return False

    def completeDeferreds(self, results, response, request, deferred_msgs):
        """A response's deferred methods have completed."""
        for i, result in enumerate(results):
            msg = deferred_msgs[i]

            if result[0] is False:
                # Invokation failed
                msg.convertFail(result[1].value)
            else:
                if msg.is_flex_msg:
                    msg.body.body = result[1]
                else:
                    msg.body = result[1]
 
        d = defer.Deferred()
        d.addCallbacks(self.encode, self.fail, errbackArgs=(request,))
        d.addCallbacks(self.finish, self.fail, callbackArgs=(request,),
            errbackArgs=(request,))
        d.callback(response) 

    def checkComplete(self, response, request):
        """Checks to determine if the response message is ready
        to be returned to the client, and defers the response if
        ready.
        """

        if hasattr(response, self.MSG_NOT_COMPLETE):
            # long-poll operation.
            # response is waiting for a message to be published.
            return

        # Check for deferred messages
        deferreds = []
        deferred_msgs = []
        for msg in response.messages:
            deferred = self.getDeferred(msg)
            if deferred is not False:
                deferreds.append(deferred)
                deferred_msgs.append(msg)

        if len(deferreds) > 0:
            dl = defer.DeferredList(deferreds)
            dl.addCallbacks(self.completeDeferreds, self.fail,
                callbackArgs=(response, request, deferred_msgs), errbackArgs=(request,))
            return

        # Message is complete, encode and return
        d = defer.Deferred()
        d.addCallbacks(self.encode, self.fail, errbackArgs=(request,))
        d.addCallbacks(self.finish, self.fail, callbackArgs=(request,), errbackArgs=(request,))
        d.callback(response)

    def finish(self, raw_response, request):
        """Send response to client when message is complete."""
        request.setHeader('Content-Type', self.CONTENT_TYPE)
        request.write(raw_response)
        request.finish()

    def fail(self, failure, request, code=500, message='Internal Server Error'):
        request.setResponseCode(500, message)
        self.finish(message, request)

    def setupPollRequest(self, packet, connection):
        """Setup a request for a long-poll operation."""

        # Set flag so self.finish
        # does not get called when
        # response packet is returned 
        # from self.invoke.
        setattr(packet.response, self.MSG_NOT_COMPLETE, True)

        request = getattr(packet, self.TWISTED_REQUEST)

        return request

    def finishPoll(self, request, packet, message, messages):
        """Finish a request that has been waiting for messages."""

        if request.finished:
            # Someone else has already called this function,
            # or twisted has finished the request for some other reason.
            return

        if isinstance(self.endpoint, AmfEndpoint):
            # Make sure messages are not encoded as an ArrayCollection
            messages = AsNoProxy(messages)
        message.response_msg.body.body = messages
 
        delattr(packet.response, self.MSG_NOT_COMPLETE)
        self.checkComplete(packet.response, request)

    def _waitForMessage(self, packet, message, connection):
        """Overridden to be non-blocking."""

        request = self.setupPollRequest(packet, connection)

        timeout_call = None
        def _notify():
            # This function gets called when a message is published,
            # or wait_interval is reached.
            if timeout_call is not None and timeout_call.active():
                # Disable time out callback
                timeout_call.cancel()

            # Disable notify function.
            connection.unSetNotifyFunc()

            # Get messages and add them
            # to the response message
            messages = self.channel_set.subscription_manager.pollConnection(connection)
            self.finishPoll(request, packet, message, messages)
        connection.setNotifyFunc(_notify)

        # Notify when wait_interval is reached
        if self.wait_interval > -1:
            timeout_call = reactor.callLater(float(self.wait_interval) / 1000, _notify)

        # Cleanup if client drops connection.
        _connectionLost = request.connectionLost
        def _connection_lost(reason):
            _connectionLost(reason)
            _notify()
        request.connectionLost = _connection_lost

        return ()

    def _pollForMessage(self, packet, message, connection):
        """Overridden to be non-blocking."""

        request = self.setupPollRequest(packet, connection)

        poller = task.LoopingCall(None)
        
        def _timeout():
            poller.stop()
            messages = self.channel_set.subscription_manager.pollConnection(connection)
            self.finishPoll(request, packet, message, messages)

        # Cleanup if client drops connection.
        _connectionLost = request.connectionLost
        def _connection_lost(reason):
            _connectionLost(reason)
            _timeout()
        request.connectionLost = _connection_lost

        if self.wait_interval > -1:
            timeout_call = reactor.callLater(float(self.wait_interval) / 1000, _timeout)
        else:
            timeout_call = None

        def _poll():
            messages = self.channel_set.subscription_manager.pollConnection(connection)
            if len(messages) > 0:
                poller.stop()
                if timeout_call is not None and timeout_call.active():
                    # Disable time out callback
                    timeout_call.cancel()

                self.finishPoll(request, packet, message, messages)
            elif connection.connected is False:
                _timeout()

        poller.f = _poll
        poller.start(float(self.poll_interval) / 1000)

        return ()

class StreamingTwistedChannel(TwistedChannel):
    """Handles streaming http connections."""

    def __init__(self, name, max_connections=-1, endpoint=None,
        timeout=1200, wait_interval=0, heart_interval=30000):
        TwistedChannel.__init__(self, name, max_connections, endpoint,
            timeout, wait_interval)

        self.heart_interval = heart_interval

    def render_POST(self, request):
        """Process an incoming AMF packet."""

        if request.getHeader('Content-Type') == self.CONTENT_TYPE:
            # Regular AMF message
            return TwistedChannel.render_POST(self, request)

        request.setHeader('Content-Type', self.CONTENT_TYPE)

        # Create streaming message command
        msg = messaging.StreamingMessage()
        msg.parseArgs(request.args)

        d = defer.Deferred()
        if msg.operation == msg.OPEN_COMMAND:
            d.addCallbacks(self.startStream, self.fail, callbackArgs=(request,), errbackArgs=(request,))
            d.callback(msg)
            return server.NOT_DONE_YET
        if msg.operation == msg.CLOSE_COMMAND:
            d.addCallbacks(self.stopStream, self.fail, callbackArgs=(request,), errbackArgs=(request,))
            d.callback(msg)
            return server.NOT_DONE_YET

    def startStream(self, msg, request):
        """Get this stream rolling!"""

        connection = self.channel_set.connection_manager.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))

        if self.channel_set.notify_connections is True:
            poller = None
        else:
            # Repeatedly poll for messages.
            poller = task.LoopingCall(None)

        # Remove notify function if client drops connection.
        _connectionLost = request.connectionLost
        def _connection_lost(reason):
            self.channel_set.disconnect(connection)
            connection.unSetNotifyFunc()
            _connectionLost(reason)
        request.connectionLost = _connection_lost

        # This function gets called when a message is published.
        def _notify():
            if connection.connected is False:
                if poller is not None:
                    poller.stop()

                connection.unSetNotifyFunc()
                msg = messaging.StreamingMessage.getDisconnectMsg()
                request.write(messaging.StreamingMessage.prepareMsg(msg, self.endpoint))
                request.finish()
                return

            msgs = self.channel_set.subscription_manager.pollConnection(connection)
            for msg in msgs:
                request.write(messaging.StreamingMessage.prepareMsg(msg, self.endpoint))
        connection.setNotifyFunc(_notify)

        if poller is not None:
            poller.f = _notify
            poller.start(float(self.poll_interval) / 1000, False) 

        # Acknowledge connection
        response = msg.acknowledge()
        response.body = connection.id
        self.sendMsg(request, response)

        # Prime request with bytes so 
        # client will start responding.
        request.write(chr(messaging.StreamingMessage.NULL_BYTE) * self.KICKSTART_BYTES)
 
        self.startBeat(connection, request)

    def sendMsg(self, request, msg):
        """Send a message to the client."""

        request.write(messaging.StreamingMessage.prepareMsg(msg, self.endpoint))

    def stopStream(self, msg, request):
        connection = self.channel_set.connection_manager.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))
        self.channel_set.disconnect(connection)

    def startBeat(self, connection, request):
        # Send out heart beat.
        looper = task.LoopingCall(None)
        def _beat():
            """Keep calling this method as long as the connection is alive."""
            if connection.connected is False:
                looper.stop()
                return
            
            request.write(chr(messaging.StreamingMessage.NULL_BYTE))

        looper.f = _beat
        looper.start(float(self.heart_interval) / 1000, False)



"""Endpoints specify how messages are encoded and decoded."""

import amfast

class Endpoint(object):
    def logRaw(self, label, raw):
        if hasattr(raw, "upper"):
            amfast.logger.debug("<%s>%s</%s>" %
                    (label, repr(raw), label))

class AmfEndpoint(Endpoint):
    """An Endpoint that can encode/decode AMF packets.

    arguments
    ==========
     * encoder - amfast.encoder.Encoder, object used to encode AMF Packets.
     * decoder - amfast.decoder.Decoder, object used to decode AMF Packets.
    """

    def __init__(self, encoder=None, decoder=None):
        if encoder is None:
            from amfast.encoder import Encoder
            encoder = Encoder()
        self.encoder = encoder

        if decoder is None:
            from amfast.decoder import Decoder
            decoder = Decoder()
        self.decoder = decoder

    def decodePacket(self, raw_packet, *args, **kwargs):
        """Decode an AMF packet."""
        if amfast.log_raw:
            self.logRaw('rawDecodePacket', raw_packet)

        return self.decoder.decode_packet(raw_packet)

    def encodePacket(self, packet):
        """Encode an AMF packet."""
        raw_packet = self.encoder.encode_packet(packet)

        if amfast.log_raw:
            self.logRaw('rawEncodePacket', raw_packet)

        return raw_packet

    def decode(self, raw_obj, amf3=None):
        """Decode an AMF object."""
        if amfast.log_raw:
            self.logRaw('rawDecodeObject', raw_obj)

        return self.decoder.decode(raw_obj, amf3)

    def encode(self, obj, amf3=None):
        """Encode an AMF object."""
        raw_obj = self.encoder.encode(obj, amf3)

        if amfast.log_raw:
            self.logRaw('rawEncodeObject', raw_obj)

        return raw_obj



# -*- coding: utf-8 -*-
import gc
import json
import re
import sys
import unittest

from guppy import hpy

from amfast.decoder import Decoder
from amfast.encoder import Encoder

class MemTestCase(unittest.TestCase):
    pattern = re.compile(r'Total size = (.+) bytes')

    def heapSize(self):
        # There's got to be a better way to do this???
        hp = hpy()
        size = str(hp.heap())
        m = self.pattern.search(size)
        return int(m.group(1))

    def _testEncoder(self, encoder, json_file):
         in_file = open(json_file)
         try:
             for idx, line in enumerate(in_file):
                 start = self.heapSize()

                 obj = json.loads(line)
                 for count in xrange(1000):
                     out = encoder.encode(obj)

                 obj = None
                 en = None
                 out = None
                 gc.collect()

                 end = self.heapSize()
                 print '%s (%s): %s - %s' % (idx, line[:10], start, end)
                 sys.stdout.flush()
         finally:
             in_file.close()

    def _testDecoder(self, decoder, encoder, json_file):
         in_file = open(json_file)
         try:
             for idx, line in enumerate(in_file):
                 if (not line.startswith('{"status"')):
                     continue

                 start = self.heapSize()

                 obj = json.loads(line)
                 for count in xrange(1000000):
                     out = encoder.encode(obj)
                     new = decoder.decode(out)

                 obj = None
                 en = None
                 out = None
                 new = None
                 gc.collect()

                 end = self.heapSize()
                 print '%s (%s): %s - %s' % (idx, line[:10], start, end)
                 sys.stdout.flush()
         finally:
             in_file.close()

         

    def testEncodeAmf0(self):
         return
         self._testEncoder(Encoder(), 'sample.log')

    def testEncodeAmf3(self):
         return
         self._testEncoder(Encoder(amf3=True), 'sample.log')

    def testDecodeAmf0(self):
         return
         self._testDecoder(Decoder(), Encoder(), 'sample.log')

    def testDecodeAmf3(self):
         return
         self._testDecoder(Decoder(amf3=True), Encoder(amf3=True), 'sample_2.log')

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(MemTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



import time
import uuid

try:
    # Load Google's special API if we're running
    # under GAE, and set flag so we know.
    import google.appengine.api.memcache as memcache
    google = True
except ImportError:
    import memcache
    google = False

class MemcacheMutex(object):
    """Blocks access to a specific key. Locking will fail if lock element is pushed out of cache."""

    def __init__(self, mc):
        self.mc = mc
        self.id = str(uuid.uuid4())
        self.acquired_locks = {}

    def acquire(self, key):
        lock_holder = self.mc.get(key)
        while lock_holder != self.id:
            if lock_holder is None:
                self.mc.set(key, self.id)
            else:
                time.sleep(0.1)
            lock_holder = self.mc.get(key)

        self.acquired_locks[key] = True

    def release(self, key, force=False):
        if key in self.acquired_locks or force is True:
            self.mc.delete(key)
            del self.acquired_locks[key]

    def releaseAll(self):
        self.mc.delete_multi(self.acquired_locks.keys())
        self.acquired_locks = {}

class MemcacheManager(object):
    KEY_SEPARATOR = '_'

    LOCK = '_lock'

    @classmethod
    def createMcClient(cls, mc_servers=("127.0.0.1:11211",), mc_debug=0):
        if google is True:
            return memcache.Client()
        else:
            return memcache.Client(mc_servers, mc_debug)

    @classmethod
    def getKeyName(cls, connection_id, attr):
        return str(cls.KEY_SEPARATOR.join((connection_id, str(attr))))

    @classmethod
    def getLockName(cls, attr):
        return str(cls.KEY_SEPARATOR.join((cls.LOCK, attr)))



# -*- coding: utf-8 -*-
import unittest

from amfast.context import EncoderContext
import amfast.encode as encode
from amfast.encoder import Encoder
import amfast.class_def as class_def

class Amf0EncoderTestCase(unittest.TestCase):
    class Spam(object):
        def __init__(self):
            self.spam = 'eggs'

    def setUp(self):
        self.class_mapper = class_def.ClassDefMapper()
        self.encoder = Encoder()

    def testFalse(self):
        self.assertEquals('\x01\x00', encode.encode(False))

    def testEncoderContext(self):
        self.assertEquals('\x01\x00', encode.encode(False, EncoderContext()))

    def testEncoderObj(self):
        self.assertEquals('\x01\x00', self.encoder.encode(False))

    def testTrue(self):
        self.assertEquals('\x01\x01', encode.encode(True))

    def testNumber(self):
        tests = {
            0: '\x00\x00\x00\x00\x00\x00\x00\x00\x00',
            0.2: '\x00\x3f\xc9\x99\x99\x99\x99\x99\x9a',
            1: '\x00\x3f\xf0\x00\x00\x00\x00\x00\x00',
            42: '\x00\x40\x45\x00\x00\x00\x00\x00\x00',
            -123: '\x00\xc0\x5e\xc0\x00\x00\x00\x00\x00',
            1.23456789: '\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b'
        }

        for number, encoding in tests.iteritems():
            self.assertEquals(encoding, encode.encode(number))

    def testString(self):
        tests = {
            '': '\x02\x00\x00',
            'hello': '\x02\x00\x05hello'
        }

        for string, encoding in tests.iteritems():
            self.assertEquals(encoding, encode.encode(string))

    def testLongString(self):
        decoded = 's' * 65537
        encoded = '\x0C\x00\x01\x00\x01' + decoded

        self.assertEquals('\x0C\x00\x01\x00\x01', encode.encode(decoded)[0:5])
        self.assertEquals(65537 + 5, len(encode.encode(decoded)))

    def testNull(self):
        self.assertEquals('\x05', encode.encode(None))

    def testTuple(self):
        decoded = (0, 1, 1.23456789)
        encoded = '\x0A\x00\x00\x00\x03' # 3 element array header
        encoded += '\x00\x00\x00\x00\x00\x00\x00\x00\x00' # element 1
        encoded += '\x00\x3f\xf0\x00\x00\x00\x00\x00\x00' #element 2
        encoded += '\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b' #element 3

        self.assertEquals(encoded, encode.encode(decoded))

    def testList(self):
        decoded = [0, 1, 1.23456789]
        encoded = '\x0A\x00\x00\x00\x03' # 3 element array header
        encoded += '\x00\x00\x00\x00\x00\x00\x00\x00\x00' # element 1
        encoded += '\x00\x3f\xf0\x00\x00\x00\x00\x00\x00' #element 2
        encoded += '\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b' #element 3

        self.assertEquals(encoded, encode.encode(decoded))

    def testCollection(self):
        from amfast.class_def.as_types import AsProxy
        decoded = [0, 1, 1.23456789]
        encoded = '\x0A\x00\x00\x00\x03' # 3 element array header
        encoded += '\x00\x00\x00\x00\x00\x00\x00\x00\x00' # element 1
        encoded += '\x00\x3f\xf0\x00\x00\x00\x00\x00\x00' #element 2
        encoded += '\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b' #element 3

        self.assertEquals(encoded, encode.encode(AsProxy(decoded)))

    def testNoCollection(self):
        from amfast.class_def.as_types import AsNoProxy
        decoded = [0, 1, 1.23456789]
        encoded = '\x0A\x00\x00\x00\x03' # 3 element array header
        encoded += '\x00\x00\x00\x00\x00\x00\x00\x00\x00' # element 1
        encoded += '\x00\x3f\xf0\x00\x00\x00\x00\x00\x00' #element 2
        encoded += '\x00\x3f\xf3\xc0\xca\x42\x83\xde\x1b' #element 3

        self.assertEquals(encoded, encode.encode(AsNoProxy(decoded)))

    def testDict(self):
        encoded = '\x03' #header
        encoded += '\x00\x04spam\x02\x00\x04eggs' #values
        encoded += '\x00\x00\t' # terminator
        self.assertEquals(encoded, encode.encode({'spam': 'eggs'}))

    def testNoProxy(self):
        from amfast.class_def.as_types import AsProxy
        encoded = '\x03' #header
        encoded += '\x00\x04spam\x02\x00\x04eggs' #values
        encoded += '\x00\x00\t' # terminator
        self.assertEquals(encoded, encode.encode(AsNoProxy({'spam': 'eggs'})))

    def testProxy(self):
        from amfast.class_def.as_types import AsProxy
        encoded = '\x03' #header
        encoded += '\x00\x04spam\x02\x00\x04eggs' #values
        encoded += '\x00\x00\t' # terminator
        self.assertEquals(encoded, encode.encode(AsProxy({'spam': 'eggs'})))

    def testNoProxy(self):
        from amfast.class_def.as_types import AsNoProxy
        encoded = '\x03' #header
        encoded += '\x00\x04spam\x02\x00\x04eggs' #values
        encoded += '\x00\x00\t' # terminator
        self.assertEquals(encoded, encode.encode(AsNoProxy({'spam': 'eggs'})))

    def testDate(self):
        import datetime
        decoded = datetime.datetime(2005, 3, 18, 1, 58, 31)
        encoded = '\x0BBp+6!\x15\x80\x00\x00\x00'
        self.assertEquals(encoded, encode.encode(decoded))

    def testXml(self):
        import xml.dom.minidom
        
        xml_str = '<?xml version="1.0" ?><test>\n            <test_me>tester</test_me>\n           </test>' # encoded XML
        decoded = xml.dom.minidom.parseString(xml_str)

        encoded = '\x0F' # XML header
        encoded += '\x00\x00\x00\x55' # String header
        encoded += xml_str

        self.assertEquals(encoded, encode.encode(decoded))

    def testAnonymousObj(self):
        decoded = self.Spam()
        decoded.spam = 'eggs'

        encoded = '\x03' #header
        encoded += '\x00\x04spam\x02\x00\x04eggs' #values
        encoded += '\x00\x00\t' # terminator
        self.assertEquals(encoded, encode.encode(decoded))

    def testDynamicObj(self):
        self.class_mapper.mapClass(class_def.DynamicClassDef(self.Spam, alias='alias.spam',
            static_attrs=(), amf3=False))

        decoded = self.Spam()
        decoded.spam = 'eggs'

        encoded = '\x10\x00\x0Aalias.spam'
        encoded += '\x00\x04spam\x02\x00\x04eggs\x00\x00\x09' # dynamic attrs

        context = EncoderContext(class_def_mapper=self.class_mapper)
        self.assertEquals(encoded, encode.encode(decoded, context))
        self.class_mapper.unmapClass(self.Spam)

    def testStaticObj(self):
        self.class_mapper.mapClass(class_def.ClassDef(self.Spam, alias='alias.spam',
            static_attrs=('spam', ), amf3=False))

        decoded = self.Spam()
        decoded.spam = 'eggs'

        encoded = '\x10\x00\x0Aalias.spam'
        encoded += '\x00\x04spam\x02\x00\x04eggs\x00\x00\x09' # dynamic attrs

        context = EncoderContext(class_def_mapper=self.class_mapper)
        self.assertEquals(encoded, encode.encode(decoded, context))
        self.class_mapper.unmapClass(self.Spam)

    def testReferences(self):
        obj = {'spam': 'eggs'}
        decoded = [obj, obj, obj]
        decoded.append(decoded)

        encoded = '\x0A\x00\x00\x00\x04' # 3 element array header
        encoded += '\x03\x00\x04spam\x02\x00\x04eggs\x00\x00\t' # obj 1
        encoded += '\x07\x00\x01' # ref to obj 1
        encoded += '\x07\x00\x01' # ref to obj 1
        encoded += '\x07\x00\x00' # circular ref

        self.assertEquals(encoded, encode.encode(decoded))

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(Amf0EncoderTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



import time
import unittest

from amfast.remoting.subscription_manager import MemorySubscriptionManager
import amfast.remoting.flex_messages as messaging

class SubscriptionTestCase(unittest.TestCase):

    class Connection(object):
        def __init__(self, **kwargs):
            for key, val in kwargs.iteritems():
                setattr(self, key, val)

    def setUp(self):
        self.manager.reset()

        self.connection_id = '95e714de-bd15-48f9-b496-8b2c5cb1af78'
        self.client_id = '95e714de-bd15-48f9-b496-8b2c5cb1af60'
        self.topic = 'test'
        self.sub_topic = 'tester'

    def makeMessage(self, topic=None, sub_topic=None):
        if topic is None:
            topic = self.topic

        if sub_topic is None:
            sub_topic = self.sub_topic

        msg = messaging.AsyncMessage(body='tester', clientId=None,
            destination=topic,
            headers={messaging.AsyncMessage.SUBTOPIC_HEADER: sub_topic},
            timeToLive=30000, timestamp=time.time() * 1000, messageId=self.client_id)
        return msg

    def testSubscribe(self):
        self.manager.subscribe(self.connection_id, self.client_id,
            self.topic, self.sub_topic)

        for connection_id in self.manager.iterSubscribers(self.topic, self.sub_topic):
            self.assertEquals(self.connection_id, connection_id)

        self.manager.unSubscribe(self.connection_id, self.client_id,
            self.topic, self.sub_topic)

        count = 0
        for subscription in self.manager.iterSubscribers(self.topic, self.sub_topic):
            count += 1
        self.assertEquals(0, count)

    def testReset(self):
        self.manager.subscribe(self.connection_id, self.client_id,
            self.topic, self.sub_topic)

        self.manager.reset()

        count = 0
        for subscription in self.manager.iterSubscribers(self.topic, self.sub_topic):
            count += 1
        self.assertEquals(0, count)

    def testDeleteConnection(self):
        self.manager.subscribe(self.connection_id, self.client_id,
            self.topic, self.sub_topic)

        count = 0
        for subscription in self.manager.iterSubscribers(self.topic, self.sub_topic):
            count += 1
        self.assertEquals(1, count)

        connection = self.Connection(id=self.connection_id)
        self.manager.deleteConnection(connection)

        count = 0
        for subscription in self.manager.iterSubscribers(self.topic, self.sub_topic):
            count += 1
        self.assertEquals(0, count)

    def testPublishAndPoll(self):
        cutoff_time = time.time() * 1000

        self.manager.subscribe(self.connection_id, self.client_id,
            self.topic, self.sub_topic)

        self.manager.publishMessage(self.makeMessage())

        msgs = self.manager.pollMessages(self.manager.getTopicKey(self.topic, self.sub_topic),
            cutoff_time, cutoff_time)

        count = 0
        for msg in msgs:
            count += 1
        self.assertEquals(1, count)

    def testPollIgnoresDifferentTopics(self):
        cutoff_time = time.time() * 1000

        self.manager.subscribe(self.connection_id, self.client_id,
            self.topic, self.sub_topic)

        self.manager.publishMessage(self.makeMessage())
        self.manager.publishMessage(self.makeMessage('different', 'topic'))

        msgs = self.manager.pollMessages(self.manager.getTopicKey(self.topic, self.sub_topic),
            cutoff_time, cutoff_time)

        count = 0
        for msg in msgs:
            count += 1
        self.assertEquals(1, count)

    def testOldMessages(self):
        current_time = time.time() * 1000

        self.manager.subscribe(self.connection_id, self.client_id,
            self.topic, self.sub_topic)

        self.manager.publishMessage(self.makeMessage())
        cutoff_time = time.time() * 1000

        msgs = self.manager.pollMessages(self.manager.getTopicKey(self.topic, self.sub_topic),
            cutoff_time, current_time)

        count = 0
        for msg in msgs:
            count += 1
        self.assertEquals(0, count)

    def testExpiredMessages(self):
        """TODO: re-write this test.

        Expired messages now are cleaned
        in different ways depending on the
        subscription manager."""
        pass  
        #cutoff_time = time.time() * 1000

        #self.manager.subscribe(self.connection_id, self.client_id,
        #    self.topic, self.sub_topic)

        #msg = self.makeMessage()
        #msg.timeToLive = 0.00001
        #self.manager.publishMessage(msg)
        #current_time = time.time() * 1000

        #msgs = self.manager.pollMessages(self.manager.getTopicKey(self.topic, self.sub_topic),
        #    cutoff_time, current_time)

        #count = 0
        #for msg in msgs:
        #    count += 1
        #self.assertEquals(0, count)

class MemoryTestCase(SubscriptionTestCase):

    def setUp(self):
        self.manager = MemorySubscriptionManager()
        SubscriptionTestCase.setUp(self)

class GaeTestCase(SubscriptionTestCase):

    def setUp(self):
        from amfast.remoting.gae_subscription_manager import GaeSubscriptionManager
        self.manager = GaeSubscriptionManager()
        SubscriptionTestCase.setUp(self)

class MemcacheTestCase(SubscriptionTestCase):

    def setUp(self):
        from amfast.remoting.memcache_subscription_manager import MemcacheSubscriptionManager
        self.manager = MemcacheSubscriptionManager()
        SubscriptionTestCase.setUp(self)

class SaTestCase(SubscriptionTestCase):

    def setUp(self):
        import sqlalchemy as sa
        from amfast.remoting.sa_subscription_manager import SaSubscriptionManager
        engine = sa.create_engine('sqlite:///sa_test_case.db', echo=False)
        metadata = sa.MetaData()

        self.manager = SaSubscriptionManager(engine, metadata)
        self.manager.createTables()
        SubscriptionTestCase.setUp(self)


def suite():
    tests = [
        unittest.TestLoader().loadTestsFromTestCase(MemoryTestCase)
    ]

    print "\n---- Optional Subscription Tests ----"
    try:
        from amfast.remoting.gae_subscription_manager import GaeSubscriptionManager
    except Exception:
        # Skip if we're not in Gae environment.
        print "Skipping GAE test."
    else:
        print "Running GAE test."
        tests.append(unittest.TestLoader().loadTestsFromTestCase(GaeTestCase))

    try:
        import sqlalchemy as sa
        from amfast.remoting.sa_subscription_manager import SaSubscriptionManager
    except Exception:
        # Skip if SQLAlchemy is not installed.
        print "Skipping SA test."
    else:
        print "Running SA test."
        tests.append(unittest.TestLoader().loadTestsFromTestCase(SaTestCase))

    try:
        from amfast.remoting.memcache_subscription_manager import MemcacheSubscriptionManager

        # Check connection
        manager = MemcacheSubscriptionManager()
        if manager.mc.set("test", True) is not True:
            print "Memcache set failed."
            raise Error("Memcache connection failed.")
        if manager.mc.get("test") != True:
            print "Memcache get failed."
            raise Error("Memcache connection failed.")
    except Exception:
        # Skip if memcache support is not installed.
        print "Skipping Memcache test."
    else:
        print "Running Memcache test."
        tests.append(unittest.TestLoader().loadTestsFromTestCase(MemcacheTestCase))

    print "--------"

    return unittest.TestSuite(tests)


if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())   



import time

import amfast
import flex_messages as messaging

class Subscription(object):
    def __init__(self, connection_id=None, client_id=None, topic=None):
        self.connection_id = connection_id
        self.client_id = client_id
        self.topic = topic

class SubscriptionManager(object):
    """Receives and publishes Producer/Consumer style messages.

    This is an abstract base class and should be implemented by a sub-class.

    attributes
    ===========
     * secure - bool, Set to True to require pulishers and subscribers to be authenticated.
     * ttl - float, Default timeToLive in milliseconds for messages that do not have the value set.
    """

    SUBTOPIC_SEPARATOR = "_;_"

    def __init__(self, secure=False, ttl=10000):
        self.secure = secure
        self.ttl = ttl

    @classmethod
    def getTopicKey(cls, topic, sub_topic=None):
        if sub_topic is not None:
            return cls.SUBTOPIC_SEPARATOR.join((topic, sub_topic))
        else:
            return topic

    @classmethod 
    def splitTopicKey(cls, topic):
        return topic.split(cls.SUBTOPIC_SEPARATOR)

    def getMessageTopicKey(self, msg):
        """Returns topic key for a message."""

        topic = msg.destination
        if hasattr(msg, 'headers') and \
            msg.headers is not None and \
            messaging.AsyncMessage.SUBTOPIC_HEADER in msg.headers:
            sub_topic = msg.headers[messaging.AsyncMessage.SUBTOPIC_HEADER]
        else:
            sub_topic = None
        return self.getTopicKey(topic, sub_topic)

    def publishMessage(self, msg):
        # Update timestamp to current server time.
        # Is this the correct thing to do???
        msg.timestamp = time.time() * 1000
        if msg.timeToLive is None or msg.timeToLive == 0:
            # Set timeToLive if it has not been pre-set.
            msg.timeToLive = self.ttl

        self.persistMessage(msg)

    def pollConnection(self, connection, soft_touch=False):
        """Retrieves all waiting messages for a specific connection.

        parameters
        ===========
         * connection - Connection, connection to poll.
         * soft_touch - boolean, True to call connection.touchSoftPolled,
             False to call connection.touchPolled. Default = False

        """

        current_time = time.time() * 1000
        polled_msgs = []
        for subscription in self.iterConnectionSubscriptions(connection):
            polled_msgs.extend([connection.personalizeMessage(\
                subscription.client_id, msg) \
                for msg in self.pollMessages(subscription.topic, \
                    connection.last_polled, current_time)])

        if soft_touch is True:
            connection.softTouchPolled()
        else:
            connection.touchPolled()
        return polled_msgs

class MemorySubscriptionManager(SubscriptionManager):
    """Stores all subscription information in memory."""

    MSG_ATTR = 'messages'
    CONNECTION_ATTR = 'connections'

    def __init__(self, secure=False, ttl=30000):
        SubscriptionManager.__init__(self, secure=secure, ttl=ttl)

        self._lock = amfast.mutex_cls()
        self.reset()

    def reset(self):
        self._topics = {}

    def _getTopicMap(self, topic):
        """Retrieves or creates a topic map."""

        topic_map = self._topics.get(topic, None)
        if topic_map is None:
            topic_map = {self.MSG_ATTR: [], self.CONNECTION_ATTR: {}}
            self._topics[topic] = topic_map
        return topic_map

    def _cleanTopicMap(self, topic, topic_map):
        """Removes un-needed subscription data for a topic."""

        if len(topic_map[self.MSG_ATTR]) == 0 and \
            len(topic_map[self.CONNECTION_ATTR]) == 0:
            del self._topics[topic]

    def subscribe(self, connection_id, client_id, topic, sub_topic=None, selector=None):
        """Subscribe a client to a topic.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
         * client_id - string, id of messaging client that is subscribing.
         * topic - string, Topic to subscribe to.
         * sub_topic - string, Sub-Topic to subscribe to. Default = None.
        """

        topic = self.getTopicKey(topic, sub_topic)
        subscription = Subscription(connection_id=connection_id,
            client_id=client_id, topic=topic)

        self._lock.acquire()
        try:
            topic_map = self._getTopicMap(topic)
            connection_map = topic_map[self.CONNECTION_ATTR].get(connection_id, None)
            if connection_map is None:
                connection_map = {}
                topic_map[self.CONNECTION_ATTR][connection_id] = connection_map

            connection_map[client_id] = subscription
        finally:
            self._lock.release()

    def unSubscribe(self, connection_id, client_id, topic, sub_topic=None):
        """Un-Subscribe a client from a topic.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
         * client_id - string, id of messaging client that is subscribing.
         * topic - string, Topic to un-subscribe from.
         * sub_topic - string, Sub-Topic to un-subscribe from. Default = None.
        """

        topic = self.getTopicKey(topic, sub_topic)

        self._lock.acquire()
        try:
            topic_map = self._topics.get(topic, None)
            if topic_map is not None:
               connection_map = topic_map[self.CONNECTION_ATTR].get(connection_id, None)
               if connection_map is not None:
                   if client_id in connection_map:
                       if len(connection_map) == 1:
                           # This is the only subscribed client,
                           # delete all connection data.
                           del topic_map[self.CONNECTION_ATTR][connection_id]
                           self._cleanTopicMap(topic, topic_map) # Clean up topic data
                       else:
                           # Delete this single subscription
                           del connection_map[client_id]
        finally:
            self._lock.release()

    def deleteConnection(self, connection):
        """Remove all subscriptions for this connection.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
        """

        for topic, topic_map in self._topics.items():
            if connection.id in topic_map[self.CONNECTION_ATTR]:
                del topic_map[self.CONNECTION_ATTR][connection.id]
                self._cleanTopicMap(topic, topic_map)

    def iterSubscribers(self, topic, sub_topic=None):
        """Iterate through Flash client ids subscribed to a specific topic."""

        topic = self.getTopicKey(topic, sub_topic)
        topic_map = self._topics.get(topic, None)
        if topic_map is None:
            return [].__iter__()

        return topic_map[self.CONNECTION_ATTR].keys().__iter__()

    def iterConnectionSubscriptions(self, connection):
        """Iterate through all Subscriptions that belong to a specific connection."""

        for topic_map in self._topics.values():
            connection_map = topic_map[self.CONNECTION_ATTR].get(connection.id, {})
            for subscription in connection_map.values():
                yield subscription

    def persistMessage(self, msg):
        """Store a message."""

        topic = self.getMessageTopicKey(msg)

        self._lock.acquire()
        try:
            topic_map = self._getTopicMap(topic)
            topic_map[self.MSG_ATTR].append(msg) 
        finally:
            self._lock.release()

    def pollMessages(self, topic, cutoff_time, current_time):
        """Retrieves all queued messages, and discards expired messages.

        arguments:
        ===========
         * topic - string, Topic to find messages for.
         * cutoff_time - float, epoch time, only messages published
             after this time will be returned.
         * current_time - float, epoch time, used to determine if a
             message is expired.
        """

        topic_map = self._topics.get(topic, None)
        if topic_map is None:
            return

        self._lock.acquire()
        try:
            msgs = topic_map[self.MSG_ATTR]
            msg_count = len(msgs)
            idx = 0
            while idx < msg_count:
                msg = msgs[idx]
                if current_time > (msg.timestamp + msg.timeToLive):
                    # Remove expired message
                    msgs.pop(idx)
                    msg_count -= 1
                else:
                    idx += 1
                    if msg.timestamp > cutoff_time:
                        yield msg
        except:
            self._lock.release()
            raise
        self._lock.release()



import os

from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, Feature

from distutils.core import setup, Extension
from distutils.command.build_ext import build_ext

from amfast import __version__

extensions = Feature(
    "AMF encoder/decoder C extension.",
    standard = True,
    optional = True,
    ext_modules = [
        Extension('amfast.encode',
            sources=[os.path.join('amfast', 'ext_src', 'encoder.c'),
                os.path.join('amfast', 'ext_src', 'amf.c')],
            depends=[os.path.join('amfast', 'ext_src', 'amf.h'),
                os.path.join('amfast', 'ext_src', 'context.h')]),
        Extension('amfast.decode',
            sources=[os.path.join('amfast', 'ext_src', 'decoder.c'),
                os.path.join('amfast', 'ext_src', 'amf.c')],
            depends=[os.path.join('amfast', 'ext_src', 'amf.h'),
                os.path.join('amfast', 'ext_src', 'context.h')]),
        Extension('amfast.buffer',
            sources=[os.path.join('amfast', 'ext_src', 'buffer.c')],
            depends=[os.path.join('amfast', 'ext_src', 'buffer.h')]),
        Extension('amfast.context',
            sources=[os.path.join('amfast', 'ext_src', 'context.c')],
            depends=[os.path.join('amfast', 'ext_src', 'context.h')])
    ])

setup(name="AmFast",
    version = __version__,
    description = "Flash remoting framework. Includes support for NetConnection, RemoteObject, IExternizeable, Flex Producer/Consumer messaging, custom type serialization, and a C-based AMF encoder/decoder.",
    url = "http://code.google.com/p/amfast/",
    author = "Dave Thompson",
    author_email = "dthomp325@gmail.com",
    maintainer = "Dave Thompson",
    maintainer_email = "dthomp325@gmail.com",
    keywords = "amf amf0 amf3 flash flex pyamf",
    platforms = ["any"],
    test_suite = "tests.suite",
    packages = ['amfast', 'amfast.class_def', 'amfast.remoting'],
    features = {'extensions': extensions},
    install_requires = {
        'uuid': 'uuid>=1.3.0'
    },
    classifiers = [
        "Programming Language :: Python :: 2.4",
        "Programming Language :: Python :: 2.5",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Utilities"
    ])



import unittest

from amfast.encode import encode
from amfast.decode import decode
from amfast.context import EncoderContext, DecoderContext
import amfast.class_def as class_def

class RoundTripTestCase(unittest.TestCase):
    class TestObject(object):
        def __init__(self):
            self.number = None
            self.test_list = ['test']
            self.sub_obj = None
            self.test_dict = {'test': 'ignore'}
            self._float = float(1.234)
            self._int = '1'
            self._str = 123

    class TestSubObject(object):
        def __init__(self):
            self.number = None

    def setUp(self):
        self.class_mapper = class_def.ClassDefMapper()

        self.class_mapper.mapClass(class_def.DynamicClassDef(self.TestObject,
            'test_complex.test', static_attrs=()))
        self.class_mapper.mapClass(class_def.DynamicClassDef(self.TestSubObject, 'test_complex.sub', ()))

    def tearDown(self):
        self.class_mapper.unmapClass(self.TestObject)
        self.class_mapper.unmapClass(self.TestSubObject)

    def buildComplex(self, max=5):
        test_objects = []

        for i in range(0, max):
            test_obj = self.TestObject()
            test_obj.number = i
            test_obj.sub_obj = self.TestSubObject()
            test_obj.sub_obj.number = i
            test_objects.append(test_obj)

        return test_objects

    def resultTest(self, decoded):
        for obj in decoded:
            self.assertEquals(self.TestObject, obj.__class__)
            self.assertEquals(self.TestSubObject, obj.sub_obj.__class__)

    def testLongStringAmf3(self):
        decoded = 's' * 65537
	encoded = encode(decoded, EncoderContext(amf3=True))
        result = decode(DecoderContext(encoded, amf3=True))

        self.assertEquals(len(decoded), len(result))
        for char in result:
            self.assertEquals('s', char)

    def testLongListAmf3(self):
        decoded = [None] * 65537
        encoded = encode(decoded, EncoderContext(amf3=True))
        result = decode(DecoderContext(encoded, amf3=True))

        self.assertEquals(len(decoded), len(result))
        for val in result:
            self.assertEquals(None, val)

    def testDictAmf3(self):
        decoded = {'spam': 'eggs'}
        encoded = encode(decoded, EncoderContext(amf3=True))
        result = decode(DecoderContext(encoded, amf3=True))

        self.assertEquals(decoded['spam'], result['spam'])

    def testLongStringAmf0(self):
        decoded = 's' * 65537
        encoded = encode(decoded)
        result = decode(encoded)

        self.assertEquals(len(decoded), len(result))
        for char in result:
            self.assertEquals('s', char)

    def testLongListAmf0(self):
        decoded = [None] * 65537
        encoded = encode(decoded)
        result = decode(encoded)

        self.assertEquals(len(decoded), len(result))
        for val in result:
            self.assertEquals(None, val)

    def testDictAmf0(self):
        decoded = {'spam': 'eggs'}
        encoded = encode(decoded)
        result = decode(encoded)

        self.assertEquals(decoded['spam'], result['spam'])

    def testComplexDictProxies(self):
        complex = {'element': 'ignore', 'objects': self.buildComplex()}
        enc_context = EncoderContext(use_collections=True, use_proxies=True,
            class_def_mapper=self.class_mapper, amf3=True)
        encoded = encode(complex, enc_context)
        decoded = decode(DecoderContext(encoded, class_def_mapper=self.class_mapper, amf3=True))
        self.resultTest(decoded['objects'])

    def testComplexDict(self):
        complex = {'element': 'ignore', 'objects': self.buildComplex()}
        enc_context = EncoderContext(use_collections=False, use_proxies=False,
            class_def_mapper=self.class_mapper, amf3=True)
        encoded = encode(complex, enc_context)
        decoded = decode(DecoderContext(encoded, class_def_mapper=self.class_mapper, amf3=True))
        self.resultTest(decoded['objects'])

    def testEncodeTypes(self):
        enc_mapper = class_def.ClassDefMapper()

        enc_mapper.mapClass(class_def.DynamicClassDef(self.TestObject,
            'test_complex.test', static_attrs=(), encode_types={
                '_float': float, '_int': int, '_str': str}))
        enc_mapper.mapClass(class_def.DynamicClassDef(self.TestSubObject, 'test_complex.sub', ()))

        complex = self.buildComplex()
        enc_context = EncoderContext(class_def_mapper=enc_mapper, amf3=True, include_private=True)
        encoded = encode(complex, enc_context)
        decoded = decode(DecoderContext(encoded, class_def_mapper=self.class_mapper, amf3=True))
        self.assertEquals('float', decoded[0]._float.__class__.__name__)
        self.assertEquals('int', decoded[0]._int.__class__.__name__)
        self.assertEquals('unicode', decoded[0]._str.__class__.__name__)

    def testDecodeTypes(self):
        dec_mapper = class_def.ClassDefMapper()

        dec_mapper.mapClass(class_def.DynamicClassDef(self.TestObject,
            'test_complex.test', static_attrs=(), decode_types={
                '_float': float, '_int': int, '_str': str}))
        dec_mapper.mapClass(class_def.DynamicClassDef(self.TestSubObject, 'test_complex.sub', ()))

        complex = self.buildComplex()
        enc_context = EncoderContext(class_def_mapper=self.class_mapper, amf3=True, include_private=True)
        encoded = encode(complex, enc_context)
        decoded = decode(DecoderContext(encoded, class_def_mapper=dec_mapper, amf3=True))
        self.assertEquals('float', decoded[0]._float.__class__.__name__)
        self.assertEquals('int', decoded[0]._int.__class__.__name__)
        self.assertEquals('str', decoded[0]._str.__class__.__name__)

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(RoundTripTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())




import memcache_manager
from subscription_manager import Subscription, SubscriptionManager
import flex_messages as messaging

class MemcacheSubscriptionManager(SubscriptionManager, memcache_manager.MemcacheManager):
    """Stores all subscription information in memcache."""

    TOPIC_ATTR = '_topics'
    MSG_ATTR = '_messages'
    CONNECTION_ATTR = '_connections'

    def __init__(self, secure=False, ttl=30000, mc_servers=['127.0.0.1:11211'], mc_debug=0):
        SubscriptionManager.__init__(self, secure=secure, ttl=ttl)

        self.mc = self.createMcClient(mc_servers, mc_debug)
        self._lock = memcache_manager.MemcacheMutex(self.mc)

    def reset(self):
        self._lock.releaseAll()

        lock_name = self.getLockName('subscription_reset')
        self._lock.acquire(lock_name)
        try:
            topics = self.mc.get(self.TOPIC_ATTR)
            if topics is not None:
                for topic in topics.iterkeys():

                    msg_key = self.getKeyName(topic, self.MSG_ATTR)
                    self.mc.delete(msg_key)

                    connection_key = self.getKeyName(topic, self.CONNECTION_ATTR)
                    connections = self.mc.get(connection_key)
                    for connection_id in connections:
                        key = self.getKeyName(connection_id, topic)
                        self.mc.delete(key)
               
                    self.mc.delete(connection_key)
                self.mc.set(self.TOPIC_ATTR, {})
        finally:
            self._lock.release(lock_name)

    def _cleanTopic(self, topic):
        """Removes un-needed subscription data for a topic."""

        msg_key = self.getKeyName(topic, self.MSG_ATTR)
        msgs = self.mc.get(msg_key)

        connection_key = self.getKeyName(topic, self.CONNECTION_ATTR)
        connections = self.mc.get(connection_key)

        if (msgs is None or len(msgs) == 0) and \
            (connections is None or len(connections) == 0):

            lock_name = self.getLockName(self.TOPIC_ATTR)
            self._lock.acquire(lock_name)
            try:
                topics = self.mc.get(self.TOPIC_ATTR)
                if topic in topics:
                    del topics[topic]
                    self.mc.set(self.TOPIC_ATTR, topics)

                if msgs is not None:
                    self.mc.delete(msg_key)

                if connections is not None:
                    self.mc.delete(connection_key)
            finally:
                self._lock.release(lock_name)

    def _createTopic(self, topic):
        lock_name = self.getLockName(self.TOPIC_ATTR)
        self._lock.acquire(lock_name)
        try:
            topics = self.mc.get(self.TOPIC_ATTR)
            if topics is None:
                topics = {}
            topic_map = topics.get(topic, None)
            if topic_map is None:
                topics[topic] = True
                self.mc.set(self.TOPIC_ATTR, topics)
        finally:
            self._lock.release(lock_name)

    def _createTopicMessageQeue(self, topic):
        key = self.getKeyName(topic, self.MSG_ATTR)
        lock_name = self.getLockName(key)
        self._lock.acquire(lock_name)
        try:
            messages = self.mc.get(key)
            if messages is None:
                self.mc.set(key, [])
        finally:
            self._lock.release(lock_name)

    def _createConnectionList(self, topic, connection_id):
        key = self.getKeyName(topic, self.CONNECTION_ATTR)
        lock_name = self.getLockName(key)
        self._lock.acquire(lock_name)
        try:
            connections = self.mc.get(key)
            if connections is None:
                connections = {}

            connections[connection_id] = True
            self.mc.set(key, connections)
        finally:
            self._lock.release(lock_name)

    def _createClientList(self, topic, connection_id, client_id, subscription):
        key = self.getKeyName(connection_id, topic)
        lock_name = self.getLockName(key)
        self._lock.acquire(lock_name)
        try:
            connection_map = self.mc.get(key)
            if connection_map is None:
                connection_map = {}

            connection_map[client_id] = subscription
            self.mc.set(key, connection_map)
        finally:
            self._lock.release(lock_name)

    def subscribe(self, connection_id, client_id, topic, sub_topic=None, selector=None):
        """Subscribe a client to a topic.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
         * client_id - string, id of messaging client that is subscribing.
         * topic - string, Topic to subscribe to.
         * sub_topic - string, Sub-Topic to subscribe to. Default = None.
        """

        topic = self.getTopicKey(topic, sub_topic)
        subscription = Subscription(connection_id=connection_id,
            client_id=client_id, topic=topic)

        self._createTopic(topic)
        self._createTopicMessageQeue(topic)
        self._createConnectionList(topic, connection_id)
        self._createClientList(topic, connection_id, client_id, subscription)

    def unSubscribe(self, connection_id, client_id, topic, sub_topic=None):
        """Un-Subscribe a client from a topic.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
         * client_id - string, id of messaging client that is subscribing.
         * topic - string, Topic to un-subscribe from.
         * sub_topic - string, Sub-Topic to un-subscribe from. Default = None.
        """

        topic = self.getTopicKey(topic, sub_topic)

        key = self.getKeyName(connection_id, topic)
        lock_name = self.getLockName(key)
        self._lock.acquire(lock_name)
        try:
            connection_map = self.mc.get(key)
            if connection_map is None:
                # Connection has already been removed.
                return
            elif len(connection_map) == 1:
                # This is the only subscribed client.
                # Delete client list.
                self.mc.delete(key)

                # Delete connection_id from subscription list
                sub_key = self.getKeyName(topic, self.CONNECTION_ATTR)
                sub_lock_name = self.getLockName(sub_key)
                self._lock.acquire(sub_lock_name)
                try:
                    connections = self.mc.get(sub_key)
                    if connection_id in connections:
                        del connections[connection_id]
                        self.mc.set(sub_key, connections)
                finally:
                    self._lock.release(sub_lock_name)

                self._cleanTopic(topic)
            else:
                # Delete single client subscription
                if client_id in connection_map:
                    del connection_map[client_id]
                    self.mc.set(key, connection_map)
        finally:
            self._lock.release(lock_name)

    def deleteConnection(self, connection):
        """Remove all subscriptions for this connection.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
        """

        # Get all subscribed topics
        topics = self.mc.get(self.TOPIC_ATTR)
        for topic in topics.iterkeys():
            # Check if connection is subscribed to a topic. 
            key = self.getKeyName(connection.id, topic)
            if self.mc.get(key) is not None:
                # Connection is subscribed to this topic
                self.mc.delete(key)
          
                # Delete connection_id from subscription list
                sub_key = self.getKeyName(topic, self.CONNECTION_ATTR)
                sub_lock_name = self.getLockName(sub_key)
                self._lock.acquire(sub_lock_name)
                try:
                    connections = self.mc.get(sub_key)
                    if connection.id in connections:
                        del connections[connection.id]
                        self.mc.set(sub_key, connections)
                    self._cleanTopic(topic)
                finally:
                    self._lock.release(sub_lock_name)

    def iterSubscribers(self, topic, sub_topic=None):
        """Iterate through Flash client ids subscribed to a specific topic."""

        topics = self.mc.get(self.TOPIC_ATTR)
        for topic in topics.iterkeys():
            key = self.getKeyName(topic, self.CONNECTION_ATTR)
            connections = self.mc.get(key)
            for connection_id in connections.iterkeys():
                yield connection_id

    def iterConnectionSubscriptions(self, connection):
        """Iterate through all Subscriptions that belong to a specific connection."""

        # Get all subscribed topics
        topics = self.mc.get(self.TOPIC_ATTR)
        for topic in topics.iterkeys():
            # Check if connection is subscribed to a topic. 
            key = self.getKeyName(connection.id, topic)
            subscriptions = self.mc.get(key)
            if subscriptions is not None:
                for subscription in subscriptions.itervalues():
                    yield subscription

    def persistMessage(self, msg):
        """Store a message."""

        topic = msg.destination
        if hasattr(msg, 'headers') and \
            msg.headers is not None and \
            messaging.AsyncMessage.SUBTOPIC_HEADER in msg.headers:
            sub_topic = msg.headers[messaging.AsyncMessage.SUBTOPIC_HEADER]
        else:
            sub_topic = None
        topic = self.getTopicKey(topic, sub_topic)

        # Remove connection data,
        # so that it is not pickled
        tmp_connection = getattr(msg, 'connection', None)
        if tmp_connection is not None:
            msg.connection = None

        key = self.getKeyName(topic, self.MSG_ATTR)
        lock_name = self.getLockName(key)
        self._lock.acquire(lock_name)
        try:
            messages = self.mc.get(key)
            if messages is None:
                messages = [] 
            messages.append(msg)
            self.mc.set(key, messages)
        finally:
            self._lock.release(lock_name)

            # Restore connection data
            if tmp_connection is not None:
                msg.connection = tmp_connection

    def pollMessages(self, topic, cutoff_time, current_time):
        """Retrieves all queued messages, and discards expired messages.

        arguments:
        ===========
         * topic - string, Topic to find messages for.
         * cutoff_time - float, epoch time, only messages published
             after this time will be returned.
         * current_time - float, epoch time, used to determine if a
             message is expired.
        """

        key = self.getKeyName(topic, self.MSG_ATTR)
        lock_name = self.getLockName(key)
        self._lock.acquire(lock_name)
        try:
            msgs = self.mc.get(key)
            if msgs is None:
                return

            msg_count = len(msgs)
            set = False
            idx = 0
            while idx < msg_count:
                msg = msgs[idx]
                if current_time > (msg.timestamp + msg.timeToLive):
                    # Remove expired message
                    msgs.pop(idx)
                    msg_count -= 1
                    set = True
                else:
                    idx += 1
                    if msg.timestamp > cutoff_time:
                        yield msg

            if set is True:
                self.mc.set(key, msgs)
        except:
            self._lock.release(lock_name)
            raise
        self._lock.release(lock_name)



import types
import threading

from django import http

import amfast
from amfast.remoting import Packet
import amfast.remoting.flex_messages as messaging
from amfast.remoting.channel import HttpChannel, ChannelError

def django_response_wrapper(func):
    '''
    A decorator which wrap a bare response to a DjangoResopnse  
    '''
    def _(channel, django_request):
        response_packet = func(channel, django_request)
        if response_packet is None:
            return http.HttpResponse(mimetype = channel.CONTENT_TYPE)
        elif type(response_packet) is types.GeneratorType:
            http_response = http.HttpResponse(content=response_packet, mimetype=channel.CONTENT_TYPE)
            return http_response
        else:
            raise ChannelError('Invalid response type.')
    return _

class DjangoChannel(HttpChannel):
    """A channel that works with Django."""

    # Attribute that holds Django's
    # request object, so that it can 
    # be accessed from a target.
    DJANGO_REQUEST = '_django_request'

    def __call__(self, http_request):
        if http_request.method != 'POST':
            return http.HttpResponseNotAllowed(['POST'])

        try:
            request_packet = self.decode(http_request.raw_post_data)
            setattr(request_packet, self.DJANGO_REQUEST, http_request)
        except amfast.AmFastError, exc:
            return http.HttpResponseBadRequest(mimetype='text/plain', content=self.getBadEncodingMsg())
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            return http.HttpResponseServerError(mimetype='text/plain', content=self.getBadServerMsg())

        try:
            response_packet = self.invoke(request_packet)
            raw_response = self.encode(response_packet)

            http_response = http.HttpResponse(mimetype=self.CONTENT_TYPE)
            http_response['Content-Length'] = str(len(raw_response))
            http_response.write(raw_response)
            return http_response
        except amfast.AmFastError, exc:
            return http.HttpResponseServerError(mimetype='text/plain', content=self.getBadServerMsg())
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            return http.HttpResponseServerError(mimetype='text/plain', content=self.getBadServerMsg())

class StreamingDjangoChannel(DjangoChannel):
    """Experimental support for streaming with Django."""

    def __init__(self, name, max_connections=-1, endpoint=None,
        wait_interval=0, heart_interval=30000):
        DjangoChannel.__init__(self, name, max_connections=max_connections,
            endpoint=endpoint, wait_interval=wait_interval)
            
        self.heart_interval = heart_interval
        
    def __call__(self, http_request):
        if http_request.META['CONTENT_TYPE'] == self.CONTENT_TYPE:
            return DjangoChannel.__call__(self, http_request)
            
        try:
            body = http_request.raw_post_data
            msg = messaging.StreamingMessage()
            msg.parseBody(body)
            #django has a well wrapped http_request object which contents all the wsgi options
            msg.parseParams(http_request.META['QUERY_STRING'])
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            
        if msg.operation == msg.OPEN_COMMAND:
            return self.startStream(msg)
            
        elif msg.operation == msg.CLOSE_COMMAND:
            return self.stopStream(msg)
            
        raise ChannelError('Http streaming operation unknown: %s' % msg.operation)
    
    @django_response_wrapper
    def startStream(self, msg):
        try: 
            connection = self.channel_set.connection_manager.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            raise ChannelError('Http streaming operation unknown: %s' % msg.operation)
        
        try:
            timer = threading.Timer(float(self.heart_interval) / 1000, self.beat, (connection, ))
            timer.daemon = True
            timer.start()

            inited = False
            event = threading.Event()
            connection.setNotifyFunc(event.set)
            poll_secs = float(self.poll_interval) / 1000
            while True:
                if connection.connected is False:
                    msg = messaging.StreamingMessage.getDisconnectMsg()
                    try:
                        yield messaging.StreamingMessage.prepareMsg(msg, self.endpoint)
                    finally:
                        # Client may have already disconnected
                        return 
                        
                if inited is False:
                    # Send acknowledge message
                    response = msg.acknowledge()
                    response.body = connection.id

                    bytes = messaging.StreamingMessage.prepareMsg(response, self.endpoint)
                    inited = True

                    bytes += chr(messaging.StreamingMessage.NULL_BYTE) * self.KICKSTART_BYTES
                    yield bytes
                    
                if self.channel_set.notify_connections is True:
                    # Block until notification of new message
                    event.wait()
                else:
                    # Block until poll_interval is reached
                    event.wait(poll_secs)
                    
                # Message has been published,
                # or it's time for a heart beat

                # Remove notify_func so that
                # New messages don't trigger event.
                connection.unSetNotifyFunc()

                msgs = self.channel_set.subscription_manager.pollConnection(connection)
                
                if len(msgs) > 0:
                    while len(msgs) > 0:
                        # Dispatch all messages to client
                        for msg in msgs:
                            try:
                                bytes = messaging.StreamingMessage.prepareMsg(msg, self.endpoint)
                            except (KeyboardInterrupt, SystemExit):
                                raise
                            except Exception, exc:
                                amfast.log_exc(exc)
                                self.channel_set.disconnect(connection)
                                break

                            try:
                                yield bytes
                                # return bytes
                            except (KeyboardInterrupt, SystemExit):
                                raise
                            except:
                                # Client has disconnected
                                self.channel_set.disconnect(connection)
                                return 
                        msgs = self.channel_set.subscription_manager.pollConnection(connection)
                else:
                    # Send heart beat
                    try:
                        yield chr(messaging.StreamingMessage.NULL_BYTE)
                    except (KeyboardInterrupt, SystemExit):
                        raise
                    except:
                        # Client has disconnected
                        self.channel_set.disconnect(connection)
                        return 

                # Create new event to trigger new messages or heart beats
                event = threading.Event()
                connection.setNotifyFunc(event.set)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            self.channel_set.disconnect(connection)
            return 
            
            
    @django_response_wrapper
    def stopStream(self, msg):
        """Stop a streaming connection."""

        connection = self.channel_set.connection_manager.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))
        connection.disconnect()
        if hasattr(connection, "notify_func") and connection.notify_func is not None:
            connection.notify_func()
    
    @django_response_wrapper
    def beat(self, connection):
        """Send a heart beat."""

        if hasattr(connection, "notify_func") and connection.notify_func is not None:
            connection.notify_func()
        else:
            return

        # Create timer for next beat
        timer = threading.Timer(float(self.heart_interval) / 1000, self.beat, (connection, ))
        timer.daemon = True
        timer.start()
